(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(require('substance'), require('texture')) :
  typeof define === 'function' && define.amd ? define(['substance', 'texture'], factory) :
  (factory(global.substance,global.texture));
}(this, (function (substance,texture) { 'use strict';

  const START = 'START';
  const END = 'END';
  const EPSILON = 'EPSILON';
  const TEXT = 'TEXT';

  class DFA {
    constructor (transitions) {
      if (!transitions || Object.keys(transitions).length === 0) {
        transitions = { START: { EPSILON: END } };
      }
      this.transitions = transitions;
    }

    consume (state, id) {
      const T = this.transitions;
      // e.g. this happens, if the state is already END
      // and more tokens are coming
      if (!T[state]) return -1
      let nextState = T[state][id];
      if (nextState !== undefined) {
        return nextState
      }
      while (T[state][EPSILON] !== undefined) {
        state = T[state][EPSILON];
        if (state === END) {
          return -1
        }
        nextState = T[state][id];
        if (nextState !== undefined) {
          return nextState
        }
      }
      return -1
    }

    canConsume (state, id) {
      let nextState = this.consume(state, id);
      return (nextState !== -1)
    }

    isFinished (state) {
      const T = this.transitions;
      if (state === 'END') return true
      // if the state is invalid
      if (!T[state]) return false
      while (T[state][EPSILON] !== undefined) {
        state = T[state][EPSILON];
        if (state === 'END') return true
      }
      return false
    }

    // Helpers to analyze

    // generates all sets of tokens, reached on all different paths
    _tokensByPath () {
      const result = [];
      const transitions = this.transitions;
      if (!transitions) return []

      // group start edges by follow state
      let first = {};
      substance.forEach(transitions[START], (to, token) => {
        if (!first[to]) first[to] = [];
        first[to].push(token);
      });

      let visited = { START: true, END: true };
      substance.forEach(first, (tokens, state) => {
        // walk all states that can be reached on this path
        // and collect all tokens
        // we consider them as potential siblings, as they
        // can co-occur at the same level
        let _siblings = {};
        tokens.forEach((t) => {
          if (t !== EPSILON) {
            _siblings[t] = true;
          }
        });
        let stack = [state];
        while (stack.length > 0) {
          let from = stack.pop();
          if (state === END) continue
          visited[from] = true;
          let T = transitions[from];
          if (!T) throw new Error(`Internal Error: no transition from state ${from}`)
          let tokens = Object.keys(T);
          for (let i = 0; i < tokens.length; i++) {
            const token = tokens[i];
            const to = T[token];
            if (!visited[to]) stack.push(to);
            if (token !== EPSILON) {
              _siblings[token] = true;
            }
          }
        }
        let _siblingTokens = Object.keys(_siblings);
        if (_siblingTokens.length > 0) {
          result.push(_siblingTokens);
        }
      });
      return result
    }
  }

  DFA.START = START;
  DFA.END = END;
  DFA.EPSILON = EPSILON;
  DFA.TEXT = TEXT;

  const START$1 = DFA.START;
  const END$1 = DFA.END;
  const EPSILON$1 = DFA.EPSILON;

  /*
    DFABuilder is essentially a graph implementation
    helping to build DFAs incrementally, by composing smaller
    sub-DFAs.
  */
  class DFABuilder {
    constructor (transitions) {
      this.transitions = transitions;
    }

    addTransition (from, to, tokens) {
      if (!this.transitions) this.transitions = {};
      if (!substance.isArray(tokens)) tokens = [tokens];
      tokens.forEach(token => _addTransition(this.transitions, from, to, token));
      return this
    }

    /*
      Creates a new DFA with all state values shifted by a given offset.
      Used when appending this DFA to another one (-> sequence)

      ```
      Expression: A,B

      Graph:  S - [A] -> E
              S - [B] -> E
              =
              S - [A] -> N - [B] -> E
      ```
    */
    append (other) {
      if (this.transitions && other.transitions) {
        let t1 = substance.cloneDeep(this.transitions);
        let t2 = substance.cloneDeep(other.transitions);
        // we need to be careful with EPSILON transitions
        // so that don't end up having a NDFA.
        let firstIsOptional = Boolean(t1[START$1][EPSILON$1]);
        let secondIsOptional = Boolean(t2[START$1][EPSILON$1]);

        if (firstIsOptional) {
          // we remove the epsilon transition from the first
          // as it would be in the way for the transformations done below
          delete t1[START$1][EPSILON$1];
        }
        // for the concatenation we insert a new state and adapt
        // the transitions of the first and second DFA to use the new state
        let newState = substance.uuid();
        // let transitions of the first going to END
        // now point to the new state
        substance.forEach(t1, (T) => {
          substance.forEach(T, (to, token) => {
            if (to === END$1) {
              T[token] = newState;
            }
          });
        });
        // If the first is optional we add transitions from
        // START to the new state
        if (firstIsOptional) {
          substance.forEach(t2[START$1], (to, token) => {
            _addTransition(t1, START$1, to, token);
          });
        }
        // for concatenation we let transitions of the second DFA
        // going from and to START now go from and to the new state
        t2[newState] = t2[START$1];
        substance.forEach(t2, (T) => {
          substance.forEach(T, (to, token) => {
            if (to === START$1) {
              T[token] = newState;
            }
          });
        });
        delete t2[START$1];
        // and now we can merge in the transitions
        substance.forEach(t2, (T, from) => {
          substance.forEach(T, (to, token) => {
            _addTransition(t1, from, to, token);
          });
        });
        // finally we add back an EPSILON transition
        // if both DFAs were optional
        if (firstIsOptional && secondIsOptional) {
          _addTransition(t1, START$1, END$1, EPSILON$1);
        }
        this.transitions = t1;
      } else if (other.transitions) {
        this.transitions = substance.cloneDeep(other.transitions);
      }
      return this
    }

    /*
      Merges to DFAs.

      Used to implement choices.

      ```
      Expression: A | B

      Graph:     - [A] -
                /       \
              S          > E
                \       /
                 - [B] -

      ```
    */
    merge (other) {
      if (this.transitions && other.transitions) {
        let t1 = this.transitions;
        let t2 = other.transitions;
        substance.forEach(t2, (T, from) => {
          substance.forEach(T, (to, token) => {
            _addTransition(t1, from, to, token);
          });
        });
      } else if (other.transitions) {
        this.transitions = substance.cloneDeep(other.transitions);
      }
      return this
    }

    /*
      Creates a new DFA with same transitions
      plus an EPSILON transition from start to END

      ```
      Expression: A?

      Graph:     - [A] -
                /       \
              S          > E
                \       /
                 -  ε  -
      ```
    */
    optional () {
      let dfa = new DFABuilder(substance.cloneDeep(this.transitions));
      if (this.transitions) {
        dfa.addTransition(START$1, END$1, EPSILON$1);
      }
      return dfa
    }

    /*
      Creates a new DFA representing (A)*

      ```
      Expression: A* = (A+)?

                         /-[A]-\
                         |     |
                          \   /
                           v /
      Graph:   S -- [A] --> 1  -- ε -->  E
                \                    /
                 \   --    ε   --   /

      ```
    */
    kleene () {
      let dfa = this.plus();
      return dfa.optional()
    }

    /*
      Creates a new DFA representing (...)+ by concatenating this
      with a kleene version: A+ = A A*

      ```
      Expression: (A)+ (sequence and reflexive edge)

                         /-[A]-\
                         |     |
                          \   /
                           v /
      Graph:  S -- [A] -->  N  -- ε -> E

      ```
    */
    plus () {
      let dfa;
      if (this.transitions) {
        let t1 = substance.cloneDeep(this.transitions);
        // there might exist an EPSILON transition already
        // which we must remove to fulfill our internal
        // assumption that there is only one EPSILON transition from
        // START going to END
        const isOptional = Boolean(t1[START$1][EPSILON$1]);
        delete t1[START$1][EPSILON$1];
        // introduce a new state
        // and let all 'ending' edges point to new state
        let newState = substance.uuid();
        substance.forEach(t1, (T) => {
          substance.forEach(T, (to, token) => {
            if (to === END$1) {
              T[token] = newState;
            }
          });
        });
        // add 'ending' EPSILON transition
        _addTransition(t1, newState, END$1, EPSILON$1);
        // copy all starting edges
        substance.forEach(t1[START$1], (to, token) => {
          _addTransition(t1, newState, to, token);
        });
        // recover 'optional'
        if (isOptional) {
          _addTransition(t1, START$1, END$1, EPSILON$1);
        }
        dfa = new DFABuilder(t1);
      } else {
        dfa = new DFABuilder(substance.cloneDeep(this.transitions));
      }
      return dfa
    }
  }

  DFABuilder.singleToken = function (token) {
    let dfa = new DFABuilder();
    dfa.addTransition(START$1, END$1, token);
    return dfa
  };

  function _addTransition (transitions, from, to, token) {
    let T = transitions[from];
    if (!T) {
      transitions[from] = T = {};
    }
    if (token === EPSILON$1 && from === START$1 && to !== END$1) {
      throw new Error('The only EPSILON transition from START must be START->END')
    }
    if (T[token] && T[token] !== to) {
      console.error('Token %s already used. Ignoring this transition.', token);
      return
      // throw new Error('Token already used in this state')
    }
    T[token] = to;
  }

  const { START: START$2, END: END$2, TEXT: TEXT$1, EPSILON: EPSILON$2 } = DFA;

  // retains the structured representation
  // and compiles a DFA for efficient processing
  class Expression {
    // TODO: why does the expression need a name?
    constructor (name, root) {
      this.name = name;
      this.root = root;

      this._initialize();
    }

    _initialize () {
      this._compile();
    }

    toString () {
      return this.root.toString()
    }

    isAllowed (tagName) {
      return Boolean(this._allowedChildren[tagName])
    }

    /*
      Some structures get compiled into a DFA, for instance.
    */
    _compile () {
      this.root._compile();
    }

    _describeError (state, token) {
      let msg = [];
      if (token !== TEXT$1) {
        if (!this.isAllowed(token)) {
          msg.push(`<${token}> is not valid in <${this.name}>\nSchema: ${this.toString()}`);
        } else {
          // otherwise just the position is wrong
          msg.push(`<${token}> is not allowed at the current position in <${this.name}>.\n${this.toString()}`);
        }
      } else {
        msg.push(`TEXT is not allowed at the current position: ${state.trace.join(',')}\n${this.toString()}`);
      }
      return msg.join('')
    }
  }

  function createExpression (name, root) {
    if (root instanceof Interleave) {
      return new InterleaveExpr(name, root)
    } else {
      return new DFAExpr(name, root)
    }
  }

  class DFAExpr extends Expression {
    getInitialState () {
      return {
        dfaState: START$2,
        errors: [],
        trace: []
      }
    }

    consume (state, token) {
      const dfa = this.dfa;
      let oldState = state.dfaState;
      let newState = dfa.consume(oldState, token);
      state.dfaState = newState;
      if (newState === -1) {
        state.errors.push({
          msg: this._describeError(state, token),
          // HACK: we want to have the element with the errors
          // but actually, here we do not know about that context
          el: state.el
        });
        return false
      } else {
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      return this.dfa.isFinished(state.dfaState)
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();
      this.dfa = new DFA(this.root.dfa.transitions);
    }

    _computeAllowedChildren () {
      this._allowedChildren = _collectAllTokensFromDFA(this.dfa);
    }

    _isValid (_tokens) {
      let state = this.getInitialState();
      for (let i = 0; i < _tokens.length; i++) {
        const token = _tokens[i];
        // Note: there might be some elements which
        // are not relevant, such as empty text nodes
        // or comments etc.
        if (!token) continue
        if (!this.consume(state, token)) {
          return false
        }
      }
      return this.isFinished(state)
    }
  }

  function _collectAllTokensFromDFA (dfa) {
    // Note: collecting all children
    const children = {};
    if (dfa.transitions) {
      substance.forEach(dfa.transitions, (T) => {
        Object.keys(T).forEach((tagName) => {
          if (tagName === EPSILON$2) return
          children[tagName] = true;
        });
      });
    }
    return children
  }

  class InterleaveExpr extends Expression {
    getInitialState () {
      const dfas = this.dfas;
      const dfaStates = new Array(dfas.length);
      dfaStates.fill(START$2);
      return {
        dfaStates,
        errors: [],
        trace: [],
        // maintain the index of the dfa which has been consumed the last token
        lastDFA: 0
      }
    }

    consume (state, token) {
      const idx = this._findNextDFA(state, token);
      if (idx < 0) {
        state.errors.push({
          msg: this._describeError(state, token)
        });
        return false
      } else {
        const dfa = this.dfas[idx];
        const oldState = state.dfaStates[idx];
        const newState = dfa.consume(oldState, token);
        state.dfaStates[idx] = newState;
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      const dfas = this.dfas;
      for (let i = 0; i < dfas.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (!dfa.isFinished(dfaState)) {
          return false
        }
      }
      return true
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();

      this.blocks = this.root.blocks;
      this.dfas = this.blocks.map(b => new DFA(b.dfa.transitions));
    }

    _computeAllowedChildren () {
      this._allowedChildren = Object.assign(...this.blocks.map((block) => {
        return _collectAllTokensFromDFA(block.dfa)
      }));
    }

    _findNextDFA (state, token) {
      console.assert(state.dfaStates.length === this.dfas.length);
      const dfas = this.dfas;
      for (let i = 0; i < state.dfaStates.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (dfa.canConsume(dfaState, token)) {
          return i
        }
      }
      return -1
    }
  }

  class Token {
    constructor (name) {
      this.name = name;
    }

    toString () {
      return this.name
    }

    _compile () {
      this.dfa = DFABuilder.singleToken(this.name);
    }
  }

  class GroupExpression {
    constructor (blocks) {
      this.blocks = blocks;
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join(this.token) + ')'
    }
  }

  /*
    (a|b|c)
  */
  class Choice extends GroupExpression {
    // copy () {
    //   return new Choice(this.blocks.map(b => b.copy()))
    // }

    // _normalize () {
    //   const blocks = this.blocks
    //   for (let i = blocks.length - 1; i >= 0; i--) {
    //     let block = blocks[i]
    //     block._normalize()
    //     // unwrap doubled Choices
    //     if (block instanceof Choice) {
    //       blocks.splice(i, 1, ...(block.blocks))
    //     }
    //   }
    // }

    _compile () {
      let dfa = new DFABuilder();
      this.blocks.forEach((block) => {
        if (block instanceof Token) {
          dfa.addTransition(START$2, END$2, block.name);
        } else if (block instanceof Interleave) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.merge(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    get token () { return Choice.token }

    static get token () { return '|' }
  }

  /*
    (a,b,c) (= ordered)
  */
  class Sequence extends GroupExpression {
    // copy () {
    //   return new Sequence(this.blocks.map(b => b.copy()))
    // }

    _compile () {
      let dfa = new DFABuilder();
      this.blocks.forEach((block) => {
        if (block instanceof Token) {
          dfa.append(DFABuilder.singleToken(block.name));
        } else if (block instanceof Interleave) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.append(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    get token () { return Sequence.token }

    static get token () { return ',' }
  }

  /*
    ~(a,b,c) (= unordered)
  */
  class Interleave extends GroupExpression {
    // copy () {
    //   return new Interleave(this.blocks.map(b => b.copy()))
    // }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join(', ') + ')[unordered]'
    }

    _normalize () {}

    _compile () {
      this.blocks.forEach(block => block._compile());
    }

    get token () { return Interleave.token }

    static get token () { return '~' }
  }

  class BlockExpression {
    constructor (block) {
      this.block = block;
    }

    toString () {
      return this.block.toString() + this.token
    }
  }

  /*
    ()?
  */
  class Optional extends BlockExpression {
    // copy () {
    //   return new Optional(this.block.copy())
    // }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.optional();
      return this.dfa
    }

    get token () { return Optional.token }

    static get token () { return '?' }
  }

  /*
    ()*
  */
  class Kleene extends BlockExpression {
    // copy () {
    //   return new Kleene(this.block.copy())
    // }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.kleene();
      return this.dfa
    }

    get token () { return Kleene.token }

    static get token () { return '*' }
  }

  /*
    ()+
  */
  class Plus extends BlockExpression {
    // copy () {
    //   return new Plus(this.block.copy())
    // }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.plus();
      return this.dfa
    }

    get token () { return Plus.token }

    static get token () { return '+' }
  }

  class ElementSchema {
    constructor (name, type, attributes, expr) {
      this.name = name;
      this.type = type;
      this.attributes = attributes;
      this.expr = expr;
    }
  }

  function _isTextNodeEmpty (el) {
    return Boolean(/^\s*$/.exec(el.textContent))
  }

  function _validateElement (elementSchema, el) {
    let errors = [];
    let valid = true;
    if (!elementSchema) {
      return {
        errors: [ {
          msg: `Unknown tag <${el.tagName}>.`,
          el
        } ],
        ok: false
      }
    }

    { // Attributes
      const res = _checkAttributes(elementSchema, el);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    // Elements
    if (elementSchema.type === 'external' || elementSchema.type === 'not-implemented') ; else {
      let res = _checkChildren(elementSchema, el);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    return {
      errors,
      ok: valid
    }
  }

  function _checkAttributes(elementSchema, el) { // eslint-disable-line
    return { ok: true }
  }

  function _checkChildren (elementSchema, el) {
    // Don't validate external nodes
    if (elementSchema.type === 'external' || elementSchema.type === 'not-implemented') {
      return true
    }
    const isText = elementSchema.type === 'text';
    const expr = elementSchema.expr;
    const state = expr.getInitialState();
    const iterator = el.getChildNodeIterator();
    let valid = true;
    let tokenCount = 0;
    while (valid && iterator.hasNext()) {
      const childEl = iterator.next();
      let token;
      if (childEl.isTextNode()) {
        // Note: skipping empty text being child node of elements
        if (_isTextNodeEmpty(childEl)) {
          continue
        } else {
          token = DFA.TEXT;
        }
      } else if (childEl.isElementNode()) {
        token = childEl.tagName;
      } else if (childEl.getNodeType() === 'cdata') {
        // CDATA elements are treated as a TEXT fragment
        token = DFA.TEXT;
      } else {
        continue
      }
      tokenCount++;
      if (!expr.consume(state, token)) {
        valid = false;
      }
    }
    // add the element to the errors
    if (state.errors.length > 0) {
      state.errors.forEach((err) => {
        err.el = el;
      });
    }
    const isFinished = expr.isFinished(state);
    if (valid && !isFinished) {
      if (isText && tokenCount === 0) ; else {
        state.errors.push({
          msg: `<${el.tagName}> is incomplete.\nSchema: ${expr.toString()}`,
          el
        });
        valid = false;
      }
    }
    if (valid) {
      state.ok = true;
    }
    return state
  }

  class XMLSchema {
    constructor (elementSchemas, startElement, publicId, dtd) {
      if (!elementSchemas[startElement]) {
        throw new Error('startElement must be a valid element.')
      }
      this._elementSchemas = {};
      this.startElement = startElement;
      this.publicId = publicId;
      this.dtd = dtd;
      // wrap schemas into ElementSchemas
      substance.forEach(elementSchemas, (spec, name) => {
        this._elementSchemas[name] = new ElementSchema(spec.name, spec.type, spec.attributes, spec.expr);
      });
    }

    getIdAttribute () {
      return 'id'
    }

    getTagNames () {
      return Object.keys(this._elementSchemas)
    }

    getDocTypeParams () {
      return [this.startElement, this.publicId, this.dtd]
    }

    getElementSchema (name) {
      return this._elementSchemas[name]
    }

    getStartElement () {
      return this.startElement
    }

    validateElement (el) {
      let tagName = el.tagName;
      let elementSchema = this.getElementSchema(tagName);
      return _validateElement(elementSchema, el)
    }
  }

  /**
   * Look up a RNG file from the current directory or a list of search directories.
   *
   * @param {*} fs
   * @param {*} rngFileName
   * @param {*} currentDir
   * @param {*} searchDirs
   */

  function validateXML (xmlSchema, dom) {
    let root = dom.find(xmlSchema.getStartElement());
    if (!root) {
      return {
        errors: [{
          msg: 'Start element is missing.',
          el: dom
        }]
      }
    } else {
      return validateElement(xmlSchema, root)
    }
  }

  function validateElement (xmlSchema, el) {
    let errors = [];
    let valid = true;
    let q = [el];
    while (q.length > 0) {
      let next = q.shift();
      let res = xmlSchema.validateElement(next);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
      if (next.isElementNode()) {
        q = q.concat(next.getChildren());
      }
    }
    return {
      errors: errors,
      ok: valid
    }
  }

  function deserializeXMLSchema (xmlSchemaInput, publicId = '', dtd = '') {
    let data;
    if (substance.isString(xmlSchemaInput)) {
      data = JSON.parse(xmlSchemaInput);
    } else {
      data = xmlSchemaInput;
    }
    let literals = data.literals;
    let schemaData = data.schema;

    function _decodeLiteral (d) {
      return literals[d]
    }

    let startElement = _decodeLiteral(schemaData[0]);
    let elementSchemas = {};

    function _decodeExpression (d) {
      if (substance.isNumber(d)) {
        return new Token(_decodeLiteral(d))
      } else if (substance.isArray(d)) {
        let type = d[0];
        let content = d[1];
        switch (type) {
          case Sequence.token:
            return new Sequence(content.map(_decodeExpression))
          case Interleave.token:
            return new Interleave(content.map(_decodeExpression))
          case Choice.token:
            return new Choice(content.map(_decodeExpression))
          case Optional.token:
            return new Optional(_decodeExpression(content))
          case Plus.token:
            return new Plus(_decodeExpression(content))
          case Kleene.token:
            return new Kleene(_decodeExpression(content))
        }
      } else {
        throw new Error('invalid data')
      }
    }

    function _decodeElementSchemaData (d) {
      let name = _decodeLiteral(d[0]);
      let type = d[1] === 't' ? 'text' : 'element';
      // TODO: at some point we gonna have more complex attribute specs
      let attributes = {};
      d[2].forEach(rank => {
        let literal = _decodeLiteral(rank);
        attributes[literal] = literal;
      });
      let expr = createExpression(name, _decodeExpression(d[3]));
      return new ElementSchema(name, type, attributes, expr)
    }

    schemaData[1].forEach(elementSchemaData => {
      let elementSchema = _decodeElementSchemaData(elementSchemaData);
      elementSchemas[elementSchema.name] = elementSchema;
    });

    let schema = new XMLSchema(elementSchemas, startElement, publicId, dtd);
    return schema
  }

  var JATSArchiving11Data = {"literals":["id","xml:lang","xref","xml:base","xlink:href","xlink:type","specific-use","style","xml:space","href","other","class","TEXT","x","content-type","definitionURL","encoding","alternatives","email","uri","ext-link","related-article","related-object","mml:math","tex-math","overline-end","roman","sans-serif","sc","strike","underline","underline-start","underline-end","ruby","bold","fixed-case","italic","monospace","overline","overline-start","sub","sup","named-content","styled-content","fn","inline-graphic","chem-struct","private-char","milestone-end","abbrev","milestone-start","inline-formula","inline-supplementary-material","hr","target","mml:ci","mml:semantics","mml:csymbol","mml:cn","mml:arcsec","mml:naturalnumbers","mml:arccsch","mml:arccsc","mml:selector","mml:complexes","mml:rationals","mml:transpose","mml:reals","mml:integers","mml:outerproduct","mml:determinant","mml:matrixrow","mml:arccoth","mml:arccot","mml:arccosh","mml:arctan","mml:arccos","mml:arcsin","mml:coth","mml:csch","mml:sech","mml:tanh","mml:cosh","mml:sinh","mml:cot","mml:csc","mml:sec","mml:tan","mml:cos","mml:sin","mml:limit","mml:product","mml:sum","mml:card","mml:prsubset","mml:subset","mml:setdiff","mml:notprsubset","mml:notsubset","mml:notin","mml:in","mml:cartesianproduct","mml:intersect","mml:union","mml:list","mml:set","mml:laplacian","mml:curl","mml:grad","mml:divergence","mml:partialdiff","mml:diff","mml:int","mml:tendsto","mml:factorof","mml:approx","mml:neq","mml:leq","mml:geq","mml:lt","mml:gt","mml:eq","mml:exists","mml:forall","mml:equivalent","mml:implies","mml:not","mml:xor","mml:or","mml:and","mml:lcm","mml:gcd","mml:times","mml:plus","mml:min","mml:max","mml:exp","mml:ceiling","mml:floor","mml:imaginary","mml:real","mml:arg","mml:conjugate","mml:abs","mml:factorial","mml:root","mml:rem","mml:power","mml:minus","mml:divide","mml:quotient","mml:compose","mml:lambda","mml:moment","mml:log","mml:ln","mml:image","mml:codomain","mml:domain","mml:ident","mml:inverse","mml:interval","mml:declare","mml:fn","mml:reln","mml:piecewise","mml:apply","mml:matrix","mml:scalarproduct","mml:vectorproduct","mml:vector","mml:mode","mml:median","mml:variance","mml:primes","mml:emptyset","mml:sdev","mml:exponentiale","mml:imaginaryi","mml:notanumber","mml:true","mml:false","mml:mean","mml:pi","mml:eulergamma","mml:infinity","mml:arctanh","mml:bind","mml:share","mml:arcsinh","mml:cerror","mml:cbytes","mml:cs","mml:arcsech","mml:malignmark","label","mml:mtable","mml:menclose","mml:mfenced","mml:mphantom","mml:mpadded","mml:merror","mml:munder","mml:mstyle","mml:mroot","mml:msqrt","mml:mfrac","mml:mrow","mml:msub","mml:maligngroup","mml:ms","mml:mspace","mml:mtext","mml:mo","mml:mn","mml:mi","mml:msubsup","mml:mover","mml:maction","mml:munderover","mml:mlongdiv","mml:mstack","mml:mmultiscripts","mml:msup","break","mathcolor","mathbackground","xlink:show","xlink:actuate","xlink:role","xlink:title","media","graphic","p","code","preformat","def-list","list","array","title","chem-struct-wrap","disp-formula","disp-formula-group","table-wrap","supplementary-material","verse-group","statement","institution","disp-quote","speech","address","fig","ack","institution-wrap","permissions","table-wrap-group","fig-group","boxed-text","position","rid","object-id","kwd-group","alt-text","abstract","align","long-desc","attrib","name","year","month","day","season","fn-group","notes","orientation","glossary","ref-list","valign","sec","etal","role","issue","caption","string-name","country","toggle","name-alternatives","aff","volume","collab","lpage","issue-title","issue-part","issue-id","dir","mathvariant","width","isbn","aff-alternatives","conf-name","conf-loc","elocation-id","supplement","mathsize","fpage","volume-series","conf-date","volume-id","mml:mglyph","count","mml:bvar","sec-meta","mml:domainofapplication","mml:condition","mml:lowlimit","mml:uplimit","string-date","page-range","conf-sponsor","color","background","publisher-loc","publisher-name","fontfamily","era","fontweight","nlm-citation","conf-acronym","trans-title","source","state","postal-code","phone","fax","mml:none","city","pub-id","char","article-title","charoff","fontstyle","collab-alternatives","element-citation","fontsize","mixed-citation","series","addr-line","contrib-group","groupalign","version","calendar","trans-source","publication-format","citation-alternatives","issn","issn-l","iso-8601-date","person-group","annotation","patent","comment","edition","date","rowalign","date-in-citation","gov","part-title","chapter-title","columnalign","size","std","mml:msgroup","bio","mml:msrow","mml:mscarries","mml:msline","type","alt","data-title","pub-id-type","rowspan","superscriptshift","funding-source","accent","given-names","award-id","subscriptshift","mimetype","mime-subtype","tr","open-access","height","mml:degree","prefix","suffix","issue-sponsor","back","product","lspace","self-uri","front","journal-id","author-comment","accentunder","contrib-id","frame","location","src","crossout","degrees","depth","floats-group","cd","seq","notation","custom-meta-group","on-behalf-of","shift","linebreak","surname","body","maxsize","longdivstyle","open","rowlines","rowspacing","selection","separators","side","stackalign","rp","rt","mml:mtd","col","tbody","subtitle","page-count","string-conf","conf-theme","conf-num","nested-kwd","compound-kwd","kwd","mml:momentabout","mml:logbase","subj-group","scope","framespacing","counts","equalrows","equalcolumns","conference","funding-group","trans-abstract","history","pub-date","author-notes","title-group","article-categories","article-id","textual-form","table","form","fence","separator","def","rspace","stretchy","symmetric","minlabelspacing","minsize","largeop","movablelimits","ext-link-type","glyph-data","lineleading","linebreakstyle","linebreakmultchar","indentalign","indentshift","indenttarget","indentalignfirst","indentshiftfirst","indentalignlast","indentshiftlast","publisher-type","publication-type","volume-issue-group","lquote","rquote","trans-title-group","price","anonymous","length","leftoverhang","rightoverhang","mslinethickness","sub-article","edge","corresp","linethickness","numalign","denomalign","bevelled","displaystyle","scriptsizemultiplier","response","contrib","alignmentscope","charalign","charspacing","close","front-stub","columnlines","columnspacing","columnspan","columnwidth","rationale","td","ref","note","designator","th","std-organization","sig-block","series-title","series-text","sig","subject","compound-subject","compound-subject-part","article","abstract-type","kwd-group-type","institution-id","ali:free_to_read","article-type","glyph-ref","mml:piece","mml:otherwise","unstructured-kwd-group","compound-kwd-part","trans-subtitle","date-type","span","copyright-statement","mml:sep","fig-count","table-count","equation-count","ref-count","symbol","word-count","colspan","custom-meta","alt-title","tfoot","app-group","app","meta-name","meta-value","headers","name-style","thead","start_date","article-meta","ali:license_ref","colgroup","copyright-year","mml:annotation-xml","mml:annotation","mml:mscarry","publisher","initials","license","table-wrap-foot","assigning-authority","arrange","copyright-holder","rb","license-p","journal-id-type","award-group","funding-statement","award-type","principal-award-recipient","principal-investigator","source-type","journal-meta","journal-title-group","journal-title","journal-subtitle","mml:mprescripts","abbrev-journal-title","mml:mlabeledtr","mml:mtr","pub-type","axis","list-type","scriptlevel","prefix-word","list-content","scriptminsize","infixlinebreakstyle","decimalpoint","continued-from","term-head","def-head","def-item","term","list-item","abbr","time-stamp","access-date","speaker","verse-line","alttext","index","closure","occurrence","nargs","base","disp-level","sec-type","object-type","object-id-type","document-type","document-id-type","document-id","source-id-type","source-id","link-type","person-group-type","style-type","ref-type","target-type","veryverythinmathspace","verythinmathspace","thinmathspace","mediummathspace","thickmathspace","verythickmathspace","veryverythickmathspace","voffset","fn-type","abbrev-type","underline-style","platforms","language-version","language","executable","actiontype","display","maxwidth","overflow","altimg","altimg-width","altimg-height","altimg-valign","order","cdgroup","mode","macros","code-version","code-type","summary","border","rules","cellspacing","cellpadding","preformat-type","baseline-shift","fig-type","alt-title-type","count-type","product-type","subj-group-type","notes-type","page","vol","related-article-type","currency","authenticated","contrib-id-type","description","collab-type","supplement-type","fontchar","fontname","format","resolution","x-size","y-size","institution-id-type","dtd-version","units","license-type","deceased","equal-contrib","contrib-type","end_date","response-type"],"schema":[531,[[535,"e",[0,3,14,6,700,564],[",",[]]],[566,"t",[0,3,14,6,564],12],[347,"e",[0,3,14,6],["+",["|",[507,250,287,298,402,372,18,279,20,44,416,280,19,2,13]]]],[507,"e",[0,3,699,499,698,697,259,6,5,4,229,230,227,228],["*",["|",[404,492,289,341,267,286,283,409,250,287,298,402,372,18,279,20,44,416,280,19,2,13]]]],[578,"t",[0,3,14,6,1],["*",["|",[12,247,253,40,41,13]]]],[545,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[568,"t",[0,3,14,6],12],[574,"e",[0,3,696,6,1,5,4,229,230,227,228],["+",["|",[566,580]]]],[580,"t",[0,3,14,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,250,238,257,240,234,251,256,232,231,235,244,243,255,241,242,353,342,344,326,385,382,390,236,237,252,248,249,246,245,491]]]],[254,"e",[0,3],[",",[["*",545],["*",568],["*",578],["*",["|",[535,574]]]]]],[338,"t",[0,3,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[287,"t",[0,3,14,259,6,1],["*",["|",[12,346,335,284,333,247,253,332,331,330,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224,195]]]],[298,"e",[0,3],["+",287]],[306,"t",[0,3,356,350,14,6,1],["*",["|",[12,270,324,269,271,268,13]]]],[300,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,346,335,284,333,247,253,332,331,330]]]],[299,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[318,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,247,253]]]],[260,"t",[0,3,379,14,6],12],[297,"t",[0,3,352,14,6],["*",["|",[12,13]]]],[354,"t",[0,3,596,352,14,6],["*",["|",[12,13]]]],[355,"t",[0,3,6],["*",["|",[12,13]]]],[281,"t",[0,3,14,413,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[293,"t",[0,3,379,14,6,1,5,4,229,230,227,228],12],[292,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[395,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[291,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[401,"t",[0,3,581,6,1],12],[280,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[490,"e",[0,3,14,6,1],[",",[328,["*",542]]]],[542,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[328,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[288,"t",[0,3,413,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[307,"t",[0,3,379,14,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[305,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[487,"e",[0,3,14,6,1],[",",[["*",288],["*",307],["?",305],["*",281],["*",293],["*",291],["*",395],["?",292]]]],[492,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[279,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[416,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,247,253]]]],[572,"e",[0,3,14],["+",[",",[322,["?",321]]]]],[322,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,247,253]]]],[321,"t",[0,3,6,1],["*",["|",[12,346,335,284,333,247,253,332,331,330,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[304,"t",[0,3,14,413,6,1],12],[290,"t",[0,3,14,6,1],12],[317,"t",[0,3,14,6,1],12],[369,"t",[0,3,695,6,1],["*",["|",[12,13]]]],[301,"t",[0,3,14,413,6],12],[353,"e",[0,3],["+",["|",[342,344,326]]]],[344,"t",[0,3,486,485,352,6,1,5,4,229,230,227,228],["*",["|",[12,52,21,22,53,316,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,195,24,23,49,48,50,42,43,358,338,367,289,341,360,327,306,300,299,318,378,362,364,270,361,18,301,279,20,304,365,247,253,297,354,355,281,293,292,291,290,269,267,286,260,317,366,359,357,336,321,322,280,271,345,369,329,370,283,302,351,328,19,349,288,307,305,268,44,54,2,40,41,13]]]],[342,"e",[0,3,486,485,352,6,1,5,4,229,230,227,228],["+",["|",[52,21,22,53,316,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,195,24,23,49,48,50,42,43,358,338,367,289,341,360,327,306,300,299,318,378,362,364,270,361,18,301,279,20,304,365,247,253,297,354,355,281,293,292,291,290,269,267,286,260,317,366,359,357,336,321,322,280,271,345,369,329,370,283,302,351,328,19,349,288,307,305,268,44,54,2,40,41,13]]]],[250,"e",[0,3,14,6,1],["*",["|",[346,335,284,333,247,253,332,331,330,18,20,19,195,13]]]],[346,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,335,284,333,247,253,332,331,330]]]],[335,"t",[0,3,14,6,1],12],[284,"t",[0,3,14,284,6,1],["*",["|",[12,13]]]],[18,"t",[0,3,14,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[333,"t",[0,3,14,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[247,"t",[0,3,14,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[534,"t",[0,3,693,14,6,1],12],[253,"e",[0,3],["*",["|",[247,534]]]],[332,"t",[0,3,14,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[331,"t",[0,3,14,6,1],12],[330,"t",[0,3,14,6,1],12],[19,"t",[0,3,14,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[302,"t",[0,3,686,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,347,239]]]],[362,"e",[0,3,543,352,356,350,6],[",",[["?",["|",[[",",[["?",270],["?",269]]],271]]],["?",268],["?",324],["?",316]]]],[270,"t",[0,3,14,6,1],12],[269,"t",[0,3,14,6,1],12],[271,"t",[0,3,14,6,1],12],[324,"t",[0,3,14,6,1],12],[268,"t",[0,3,14,356,350,6,1],12],[316,"t",[0,3,356,350,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,270,324,269,271,268]]]],[341,"e",[0,3],["+",289]],[289,"t",[0,3,685,551,6,1,5,4,229,230,227,228],["*",["|",[12,52,21,22,346,335,284,333,247,253,332,331,330,53,224,347,250,287,298,402,372,18,279,20,44,416,280,19,2,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,40,41,13]]]],[404,"t",[0,3,683,682,14,6,1],12],[286,"e",[0,3],["+",["|",[267,283]]]],[267,"e",[0,3,14,562,6,1],[",",[["|",[[",",[419,["?",384]]],384]],["?",393],["?",394]]]],[283,"t",[0,3,14,562,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,409,384,393,419,394]]]],[409,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[384,"t",[0,3,573],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[419,"t",[0,3,573],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[393,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[394,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[20,"t",[0,3,473,576,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[266,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[465,"e",[0,3,259,6,1],["+",233]],[195,"t",[0,3,377,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[491,"t",[0,3,681,14,6,1],["*",["|",[12,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33]]]],[239,"t",[0,3,14,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224,353,342,344,326]]]],[21,"t",[0,3,680,473,679,678,281,301,401,581,6,1,5,4,229,230,227,228],["*",["|",[12,52,21,22,53,224,316,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,401,195,24,23,49,48,50,42,43,358,338,367,289,341,360,327,306,300,299,318,378,362,364,270,361,18,301,279,20,304,365,247,253,297,354,355,281,293,292,291,290,269,267,286,260,317,366,359,357,336,321,322,280,271,345,369,329,370,283,302,351,328,19,349,288,307,305,268,44,54,2,40,41,13]]]],[524,"t",[0,3,259,14,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224,507,232,231,287,298,279,280,527]]]],[527,"t",[0,3,259,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224,232,231]]]],[252,"e",[0,3,6,14,1],[",",[["?",311],["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]]]]],[372,"e",[0,3,259,14,6,1,5,4,229,230,227,228],[",",[["?",311],["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]]]]],[273,"e",[0,3,677,6,1],[",",[["?",311],["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]]]]],[262,"t",[0,3,14,6,1],12],[265,"t",[0,3,14,6,1,5,4,229,230,227,228],["*",["|",[12,13]]]],[415,"e",[0,3],["+",554]],[554,"e",[0,3,6,1,5,4,229,230,227,228],[",",[559,560]]],[559,"t",[0,3],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[560,"t",[0,3],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[17,"e",[0,3],["+",["|",[238,46,234,232,45,52,231,235,47,244,461,460,24,23]]]],[460,"t",[0,3,6,1],["*",["|",[12,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,45,47,195,24,23,42,43,40,41]]]],[13,"t",[0,3,14,8,6,1],["*",["|",[12,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,42,43,40,41,13]]]],[565,"e",[0,3],[",",[["*",459],["?",458],["?",457],["*",["|",[347,287,298,13]]],["?",456],["*",455],["*",288],["*",307],["?",305],["*",281],["*",293],["*",291],["*",395],["?",292],["*",487],["*",297],["?",302],["?",["|",[[",",[["?",[",",[304,["?",290]]]],["?",317]]],301]]],["*",["|",[18,20,19,397,244]]],["?",454],["?",254],["*",399],["*",["|",[21,22]]],["*",263],["*",453],["*",261],["*",452],["*",451],["?",448],["?",415]]]],[459,"t",[0,3,379,6],12],[458,"e",[0,3],[",",[["*",445],["*",525],["*",526]]]],[445,"e",[0,3,676,6,1],[",",[["+",["|",[528,529]]],["*",445]]]],[528,"t",[0,3,14],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[529,"e",[0,3,14],["+",530]],[530,"t",[0,3,14],["*",["|",[12,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,42,43,40,41]]]],[525,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[526,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[456,"e",[0,3,259,6],[",",[["?",195],["?",239],["+",["|",[499,44,233,13]]]]]],[397,"t",[0,3,675,6,1,5,4,229,230,227,228],["*",["|",[12,52,21,22,53,224,316,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,195,24,23,49,48,50,42,43,491,358,338,367,289,341,360,327,306,300,299,318,378,362,364,270,361,18,301,279,20,304,365,247,253,297,354,355,281,293,292,291,290,269,267,286,260,317,366,359,357,336,321,322,280,271,345,369,329,370,283,302,351,328,19,349,288,307,305,268,44,54,2,40,41,13]]]],[454,"t",[0,3],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224,362,316]]]],[399,"t",[0,3,14,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[263,"e",[0,3,532,6,1],[",",[["*",260],["?",311],["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]]]]],[453,"e",[0,3,532,6,1],[",",[["*",260],["?",311],["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]]]]],[261,"e",[0,3,533,6,1],[",",[["?",195],["?",239],["|",[["+",["|",[442,441,440,13]]],["*",540]]]]]],[442,"t",[0,3,14],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[441,"e",[0,3,14],["+",541]],[541,"t",[0,3,14],["*",["|",[12,224,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[440,"e",[0,3,14],[",",[["+",["|",[442,441]]],["*",440]]]],[540,"t",[0,3,533,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[499,"t",[0,3,14,6,1],["*",["|",[12,346,335,284,333,247,253,332,331,330,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224,195]]]],[455,"e",[0,3,596,352,543,356,350,1],["*",["|",[270,324,269,271,268,316,13]]]],[451,"e",[0,3,14,6,1,5,4,229,230,227,228],["*",["|",[306,299,439,300,318,438,327,437,13]]]],[327,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[439,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[438,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[437,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,306,299,439,300,318,438,327,437]]]],[448,"e",[0,3],[",",[["*",309],["?",547],["?",548],["?",549],["?",550],["?",436],["?",552]]]],[309,"e",[0,3,674,309],[",",[]]],[549,"e",[0,3,309],[",",[]]],[547,"e",[0,3,309],[",",[]]],[548,"e",[0,3,309],[",",[]]],[550,"e",[0,3,309],[",",[]]],[436,"e",[0,3,309],[",",[]]],[552,"e",[0,3,309],[",",[]]],[457,"e",[0,3],[",",[338,["*",435],["*",490],["*",555],["?",272]]]],[435,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[555,"t",[0,3,673,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[402,"e",[0,3,14,6,1],[",",[["?",239],["+",233]]]],[557,"e",[0,3,14,6,1],[",",[["?",195],["?",239],["*",263],["*",261],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",["|",[558,276]]]]]],[558,"e",[0,3,14,6,1],[",",[["?",311],["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]],["?",254]]]],[272,"e",[0,3,14,6,1],[",",[["?",195],["?",239],["+",["|",[44,13]]]]]],[275,"e",[0,3,14,6,1],[",",[["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",275]]]],[238,"e",[0,3,14,6,274,1],[",",[["?",195],["*",["|",[262,265,18,20,19]]],["|",[["*",["|",[17,232,231]]],434]],["*",["|",[266,254]]]]]],[257,"e",[0,3,258,274,6,1,14],[",",[["*",260],["?",311],["?",195],["?",282],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]],["*",["|",[266,254]]]]]],[240,"e",[0,3,258,274,6,1,14],[",",[["*",260],["?",195],["?",282],["*",263],["*",261],["*",["|",[262,265,18,20,19]]],["+",["|",[17,46,234,232,231,235,460]]],["*",["|",[266,254]]]]]],[46,"t",[0,3,14,6,1,5,4,229,230,227,228],["*",["|",[12,262,265,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224,260,195,236,237,238,234,232,231,235]]]],[256,"e",[0,3,258,274,6,1,14],[",",[["?",195],["?",282],["*",263],["*",261],["*",["|",[262,265,18,20,19]]],["*",["|",[251,17,232,231]]]]]],[251,"e",[0,3,258,274,6,1,672],[",",[["*",260],["?",195],["*",282],["*",263],["*",261],["*",["|",[262,265,18,20,19]]],["*",["|",[241,242,240,248,249,246,245,243,233,236,237,17,238,234,232,231,235]]],["*",["|",[266,254]]]]]],[282,"e",[0,3,14,6,1,7],[",",[["?",239],["*",233]]]],[232,"e",[0,3,258,274,6,1,14,388,387,5,4,229,230,227,228],["*",["|",[262,265,263,18,20,19,282,260,261,195,266,254]]]],[231,"e",[0,3,258,274,6,1,14,387,388,5,4,229,230,227,228],["*",["|",[262,265,263,18,20,19,282,260,261,195,266,254]]]],[45,"e",[0,3,14,6,671,387,388,1,5,4,229,230,227,228],["?",262]],[235,"t",[0,3,258,274,6,1,670,8],["*",["|",[12,262,265,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,266,254]]]],[234,"t",[0,3,664,663,650,649,648,647,258,274,6,1,8],["*",["|",[12,18,20,19,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,49,48,50,42,43,44,54,2,40,41]]]],[244,"e",[0,3,258,274,6,1,14,387,388,5,4,229,230,227,228],[",",[["*",260],["?",195],["*",282],["*",263],["*",261],["*",["|",[262,265,18,20,19]]],["*",["|",[241,242,240,248,249,246,245,243,233,236,237,17,238,234,232,231,235]]],["*",["|",[266,254]]]]]],[255,"e",[0,3,258,274,6,1,14],[",",[["?",195],["?",282],["*",263],["*",261],["*",["|",[262,265,18,20,19]]],["+",243]]]],[243,"e",[0,3,258,274,6,1,14],[",",[["*",260],["?",195],["?",282],["*",263],["*",261],["*",["|",[262,265,18,20,19]]],["*",["|",[248,249,246,245,236,237,17,240,234,232,231,235,461]]],["*",["|",[575,266,254]]]]]],[575,"e",[0,3],[",",[["?",195],["?",239],["+",["|",[233,272,44,266,254,13]]]]]],[53,"e",[0,3],[",",[]]],[224,"e",[0,3],[",",[]]],[34,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[35,"t",[0,3,14,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[36,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[37,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[26,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[27,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[28,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[38,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[29,"t",[0,3,285,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[40,"t",[0,3,577,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[41,"t",[0,3,577,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[30,"t",[0,3,285,646,6],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[39,"e",[0,3,6],[",",[]]],[25,"e",[0,3,259,6],[",",[]]],[31,"e",[0,3,6],[",",[]]],[32,"e",[0,3,259,6],[",",[]]],[33,"e",[0,3,14,6],[",",[579,["|",[431,[",",[430,431,430]]]]]]],[579,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[431,"t",[0,3,14,6,1],12],[430,"t",[0,3],12],[452,"e",[0,3,6,1],[",",[["*",582],["*",583],["*",390]]]],[583,"t",[0,3,259,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,385,382,390]]]],[390,"e",[0,3,6,1],["+",233]],[582,"e",[0,3,259,584,6,1,5,4,229,230,227,228],[",",[["*",382],["*",385],["*",585],["*",586]]]],[382,"t",[0,3,259,587,284,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,247,253]]]],[385,"t",[0,3,259,584,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[585,"t",[0,3,6,1],["*",["|",[12,404,267,286,247,253,283]]]],[586,"t",[0,3,6,1],["*",["|",[12,404,267,286,283]]]],[588,"e",[0,3],[",",[["*",401],["*",589],["*",["|",[347,287,298]]],["*",354],["?",355],["*",297],["?",572],["*",273],["*",399],["?",415]]]],[589,"e",[0,3,14],[",",[["*",590],["*",591],["*",490],["*",593]]]],[590,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[591,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[593,"t",[0,3,645,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[44,"e",[0,3,551,644,6,1],[",",[["?",195],["+",233]]]],[54,"t",[0,3,635,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[2,"t",[0,3,634,377,259,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[52,"t",[0,3,14,387,388,6,1,5,4,229,230,227,228],["*",["|",[12,262,265,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[236,"e",[0,3,598,600,601,605,6,1],[",",[["?",195],["?",239],["?",606],["?",607],["*",["|",[608,13]]],["*",236]]]],[606,"t",[0,3],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[607,"t",[0,3],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[608,"e",[0,3,6,1],[",",[["?",195],["*",609],["*",["|",[465,13]]]]]],[609,"t",[0,3,259,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,241,242,240,238,234,232,231,235]]]],[237,"e",[0,3,598,600,601,605,6,1],[",",[["?",195],["?",239],["+",["|",[610,13]]]]]],[610,"e",[0,3,6,1],[",",[["?",195],["?",239],["+",["|",[233,236,237]]]]]],[51,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[241,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,263,262,265,224,266,254,261,195,238,234,232,231,235]]]],[242,"e",[0,3,14,6,1],[",",[["?",195],["?",282],["*",263],["*",261],["*",["|",[262,265,18,20,19]]],["*",["|",[241,242]]]]]],[24,"t",[0,3,14,6,414,349],12],[326,"e",[0,3,486,485,352,6,1,5,4,229,230,227,228],[",",[["*",["|",[357,289]]],["*",["|",[338,328]]],["?",329],["?",359],["?",351],["?",268],["|",[[",",[["?",269],["?",270],["?",612]]],["?",271]]],["?",613],["?",288],["?",361],["?",299],["?",306],["?",300],["*",["|",[281,302]]],["?",321],["?",322],["*",[",",[["?",304],["?",290]]]],["?",436],["?",345],["*",360],["*",336],["?",358]]]],[233,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,250,238,257,240,234,251,256,232,231,235,244,243,255,241,242,353,342,344,326,385,382,390,236,237,252,248,249,246,245]]]],[248,"e",[0,3,14,6,1],[",",[["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",["|",[266,254]]]]]],[249,"e",[0,3,14,6,1],[",",[614,["+",233]]]],[614,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,409,384,393,419,394]]]],[246,"e",[0,3,14,6,1],[",",[["?",195],["?",239],["*",263],["*",261],["+",["|",[233,246]]],["*",["|",[266,254]]]]]],[245,"e",[0,3,14,6,1],[",",[["?",195],["?",239],["?",435],["+",["|",[615,245]]],["*",["|",[266,254]]]]]],[615,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[49,"t",[0,3,14,6,377,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,465]]]],[50,"e",[0,3,259,14,517,6,1],[",",[]]],[48,"e",[0,3,259,14,517,6,1],[",",[]]],[42,"t",[0,3,259,377,14,6,1,5,4,229,230,227,228],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,250,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,252,248,249,246,245]]]],[43,"t",[0,3,285,7,633,377,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,250,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,252,248,249,246,245]]]],[276,"e",[0,3,14,6,1],[",",[["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13,519]]],["*",276]]]],[519,"e",[0,3,14,6,1],[",",[["?",195],["+",["|",[353,342,344,326,520,13]]]]]],[520,"e",[0,3,14,6,1],[",",[["?",195],["+",["|",[233,397]]]]]],[613,"t",[0,3,356,350,14,6],["*",["|",[12,270,324,269,271,268,13]]]],[358,"e",[0,3,14,6,1],["+",233]],[367,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[360,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[378,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,42,43,40,41]]]],[364,"t",[0,3,356,350,14,6,1],["*",["|",[12,270,324,269,271,268,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,13]]]],[361,"t",[0,3,521,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[365,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[366,"t",[0,3,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[359,"t",[0,3,14,284,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[357,"t",[0,3,632,6,1],["*",["|",[12,492,289,341,267,286,283,287,298,279,280,13]]]],[336,"t",[0,3,379,576,6,5,4,229,230,227,228],12],[345,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[370,"t",[0,3,14,6,1],["*",["|",[12,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,42,43,270,269,336,329,523,268,40,41]]]],[523,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,247,253,224]]]],[329,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[612,"t",[0,3,14,6],["*",["|",[12,270,324,269,271,268,13]]]],[351,"t",[0,3,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,224]]]],[349,"t",[0,3,521,14,6,1],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13]]]],[22,"t",[0,3,631,473,630,629,587,628,627,626,260,625,624,14,6,1,5,4,229,230,227,228],["*",["|",[12,52,21,22,53,224,316,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,195,24,23,49,48,50,42,43,358,338,367,289,341,360,327,306,300,299,318,378,362,364,270,361,18,301,279,20,304,365,247,253,297,354,355,281,293,292,291,290,269,267,286,260,317,366,359,357,336,321,322,280,271,345,369,329,370,283,302,351,328,19,349,288,307,305,268,44,54,2,40,41,13]]]],[411,"e",[0,3],["*",["|",[17,257,240,234,251,256,232,231,235,244,243,255]]]],[278,"e",[0,3,1,623,622,6],[",",[["?",311],["?",195],["?",239],["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["*",["|",[273,272,275,276]]]]]],[311,"e",[0,3],[",",[["*",347],["*",263],["*",261],["?",254]]]],[58,"t",[4,5,1,8,0,2,11,7,9,10,16,15,376,621],["*",["|",[12,308,546,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[55,"t",[4,5,1,8,0,2,11,7,9,10,16,15,376],["*",["|",[12,308,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[57,"t",[4,5,1,8,0,2,11,7,9,10,16,15,376,412],["*",["|",[12,308,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[166,"e",[4,5,1,8,0,2,11,7,9,10],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]],["*",310],["*",["|",[["*",["|",[312,313,[",",[314,["?",315]]]]]],392,443,444]]],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[187,"e",[4,5,1,8,0,2,11,7,9,10],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]],["*",310],["*",["|",[["*",["|",[312,313,[",",[314,["?",315]]]]]],392,443,444]]],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[188,"e",[4,5,1,8,0,2,11,7,9,10,407],[",",[]]],[190,"e",[4,5,1,8,0,2,11,7,9,10],[",",[57,["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[191,"t",[4,5,1,8,0,2,11,7,9,10,16,15],12],[192,"t",[4,5,1,8,0,2,11,7,9,10,16,15],12],[310,"e",[],["|",[[",",[392,["|",[55,56]]]],[",",[["|",[55,56]],["?",392]]]]]],[546,"e",[],[",",[]]],[312,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[313,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[315,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[314,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[392,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[443,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[444,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[165,"e",[4,5,1,8,0,2,11,7,9,10,16,15],["*",["|",[538,539]]]],[538,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]],[539,"e",[4,5,1,8,0,2,11,7,9,10,16,15],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[164,"e",[],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]],[163,"e",[],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]],[162,"e",[376,446,620,619,16,15],["+",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]],[161,"e",[4,5,1,8,0,2,11,7,9,10,16,15,618],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]],[160,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[159,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[158,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[157,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[156,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[155,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[154,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[153,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[152,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[["*",310],["*",["|",[312,313,[",",[314,["?",315]]]]]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]],[151,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[150,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[149,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[148,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[147,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[146,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[145,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[144,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[143,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[142,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[141,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[140,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[139,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[138,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[137,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[136,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[135,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[134,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[133,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[132,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[131,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[130,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[129,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[128,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[127,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[126,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[125,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[124,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[123,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[122,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[121,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[120,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[119,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[118,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[117,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[116,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[115,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[114,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[113,"e",[4,5,1,8,0,2,11,7,9,10,16,15,376],[",",[]]],[112,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[111,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[110,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[109,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[108,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[107,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[106,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[105,"e",[4,5,1,8,0,2,11,7,9,10,16,15,376],[",",[["*",["*",310]],["*",["*",["|",[312,313,[",",[314,["?",315]]]]]]],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[104,"e",[4,5,1,8,0,2,11,7,9,10,16,15,659],[",",[["*",["*",310]],["*",["*",["|",[312,313,[",",[314,["?",315]]]]]]],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[103,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[102,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[101,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[100,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[99,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[98,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[97,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[96,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[95,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[94,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[93,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[92,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[91,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[90,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[89,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[88,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[87,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[86,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[85,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[84,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[83,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[82,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[81,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[80,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[79,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[78,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[77,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[76,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[75,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[74,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[73,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[72,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[62,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[61,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[59,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[193,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[189,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[186,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[182,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[176,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[173,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[172,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[171,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[170,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[["*",310],["*",["|",[312,313,[",",[314,["?",315]]]]]],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[167,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[["*",310],["*",["|",[312,313,[",",[314,["?",315]]]]]],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[71,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[["*",310],["*",["|",[312,313,[",",[314,["?",315]]]]]],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192]]]]]],[70,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[66,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[63,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[169,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[168,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[69,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[68,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[67,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[65,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[60,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[64,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[174,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[175,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[177,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[178,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[179,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[180,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[181,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[183,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[184,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[185,"e",[4,5,1,8,0,2,11,7,9,10,16,15],[",",[]]],[215,"t",[4,5,1,8,0,2,11,7,9,10,225,226,295,303,294,323,325,340,343,319,320],["*",["|",[12,308,194]]]],[214,"t",[4,5,1,8,0,2,11,7,9,10,225,226,295,303,294,323,325,340,343,319,320],["*",["|",[12,308,194]]]],[213,"t",[4,5,1,8,0,2,11,7,9,10,225,226,295,303,294,323,325,340,343,319,320,462,463,464,398,466,467,468,421,470,471,472,383,418,475,476,477,478,479,480,481,482,483,484],["*",["|",[12,308,194]]]],[212,"t",[4,5,1,8,0,2,11,7,9,10,225,226,295,303,294,323,325,340,343,319,320],["*",["|",[12,308,194]]]],[211,"e",[4,5,1,8,0,2,11,7,9,10,225,226,295,303,294,323,325,340,343,319,320,296,391,410,418],[",",[]]],[210,"t",[4,5,1,8,0,2,11,7,9,10,225,226,295,303,294,323,325,340,343,319,320,488,489],["*",["|",[12,308,194]]]],[308,"e",[4,5,1,8,0,2,11,7,9,10,225,226,407,296,391,277,377,617,295,303,323,325,340,343,319,320],[",",[]]],[375,"e",[4,5,1,8,0,2,11,7,9,10,225,226,258,493,494,495,496],[",",[]]],[334,"e",[4,5,1,8,0,2,11,7,9,10,225,226],[",",[]]],[592,"e",[4,5,1,8,0,2,11,7,9,10,225,226],[",",[]]],[194,"e",[4,5,1,8,0,2,11,7,9,10,225,226,498],[",",[]]],[209,"e",[4,5,1,8,0,2,11,7,9,10,225,226,348],[",",[]]],[207,"e",[4,5,1,8,0,2,11,7,9,10,225,226,294],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[206,"e",[4,5,1,8,0,2,11,7,9,10,225,226,500,501,502,503],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[205,"e",[4,5,1,8,0,2,11,7,9,10,225,226],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[204,"e",[4,5,1,8,0,2,11,7,9,10,225,226],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[203,"e",[4,5,1,8,0,2,11,7,9,10,225,226,599,504,505,602,603,604,383,403,264,508,503,509,510,511,368,513,514,515,516,408,502,410,294,498,450,449,463,462,405,447,348,391,478,481,483,479,482,484,480,471,494,493,418,477,476,475,500,406,422,488,398,303,295,421,469,470,472,496,414,501,423,258,495,363,424,425,380,489,466,426,464,427,417,428,429,467,386,381,468,277,296,323,325,340,343,319,320,636,637,638,639,640,641,642],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[201,"e",[4,5,1,8,0,2,11,7,9,10,225,226],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[200,"e",[4,5,1,8,0,2,11,7,9,10,225,226,391,410,296,398,643],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[199,"e",[4,5,1,8,0,2,11,7,9,10,225,226],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[198,"e",[4,5,1,8,0,2,11,7,9,10,225,226,423,511,427],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[197,"e",[4,5,1,8,0,2,11,7,9,10,225,226,414],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[208,"e",[4,5,1,8,0,2,11,7,9,10,225,226,386],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[223,"e",[4,5,1,8,0,2,11,7,9,10,225,226,381],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[216,"e",[4,5,1,8,0,2,11,7,9,10,225,226,386,381],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[202,"e",[4,5,1,8,0,2,11,7,9,10,225,226,403,264],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[217,"e",[4,5,1,8,0,2,11,7,9,10,225,226,383,264],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[219,"e",[4,5,1,8,0,2,11,7,9,10,225,226,383,403,264],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]]],[222,"e",[4,5,1,8,0,2,11,7,9,10,225,226,386,381],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["*",[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,334]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,334]]]]],["?",[",",[592,["*",[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,334]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,334]]]]]]]]]]],[196,"e",[4,5,1,8,0,2,11,7,9,10,225,226,264,363,368,348,508,516,296,425,514,424,513,405,447,449,450,504,428,469],["*",["|",[595,594]]]],[594,"e",[4,5,1,8,0,2,11,7,9,10,225,226,363,368,348],["+",432]],[595,"e",[4,5,1,8,0,2,11,7,9,10,225,226,363,368,348],["*",432]],[432,"e",[4,5,1,8,0,2,11,7,9,10,225,226,380,515,363,368,348],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[221,"e",[4,5,1,8,0,2,11,7,9,10,225,226,264,429,509,510],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,374,375,373,371]]]],[220,"e",[4,5,1,8,0,2,11,7,9,10,225,226,258,417,422],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,374,375,373,371]],["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,374,375,373,371]],["+",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,374,375,373,371]]]]]],[371,"e",[4,5,1,8,0,2,11,7,9,10,225,226,258,417],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,374,375,373,371]]]],[373,"e",[4,5,1,8,0,2,11,7,9,10,225,226,258],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,334]]]],[374,"e",[4,5,1,8,0,2,11,7,9,10,225,226,258,406,408,505],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,334,571]]]],[571,"e",[4,5,1,8,0,2,11,7,9,10,225,226,406,408],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218,334]]]],[218,"e",[4,5,1,8,0,2,11,7,9,10,225,226,651,426],["+",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[23,"e",[4,5,1,8,0,2,11,7,9,10,652,653,654,655,656,657,658,616,660,661,662,225,226,599,504,505,602,603,604,383,403,264,508,503,509,510,511,368,513,514,515,516,408,502,410,294,498,450,449,463,462,405,447,348,391,478,481,483,479,482,484,480,471,494,493,418,477,476,475,500,406,422,488,398,303,295,421,469,470,472,496,414,501,423,258,495,363,424,425,380,489,466,426,464,427,417,428,429,467,386,381,468,277,296],["*",["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]]]],[570,"t",[4,5,1,8,0,2,11,7,9,10,412,267,16,15,407],12],[569,"e",[4,5,1,8,0,2,11,7,9,10,412,267,16,15,407],["+",233]],[56,"e",[4,5,1,8,0,2,11,7,9,10,16,15,412,267],[",",[["|",[165,164,163,162,161,160,159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,139,138,137,136,135,134,133,132,131,130,129,128,127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112,111,110,109,108,107,106,105,104,103,102,101,100,99,98,97,96,95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,79,78,77,76,75,74,73,72,62,61,59,193,189,186,182,176,173,172,171,170,167,71,70,66,63,169,168,69,68,67,65,60,64,174,175,177,178,179,180,181,183,184,185,56,58,55,57,166,187,188,190,191,192,215,214,213,212,211,210,209,194,207,206,205,204,203,201,200,199,198,197,208,223,216,202,217,219,222,196,221,220,218]],["*",["|",[570,569]]]]]],[461,"e",[0,3,14,7,665,296,666,405,667,668,669,6],[",",[["|",[["*",433],["*",567]]],["|",[[",",[["?",563],["?",556],["+",434]]],["+",389]]]]]],[563,"e",[0,3,14,7,264,337,339,277],["+",389]],[556,"e",[0,3,14,7,264,337,339,277],["+",389]],[434,"e",[0,3,14,7,264,337,339,277],["+",389]],[567,"e",[0,3,14,7,544,296,264,337,339,277],["*",433]],[433,"e",[0,3,14,7,544,296,264,337,339,277],[",",[]]],[389,"e",[0,3,14,7,264,337,339,277],["+",["|",[522,518]]]],[522,"t",[0,3,14,7,611,597,561,446,380,553,264,337,339,277],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,241,242,224,240,353,342,344,326,236,237,233,238,234,232,231,235]]]],[518,"t",[0,3,14,7,611,597,561,446,380,553,264,337,339,277],["*",["|",[12,18,20,19,52,21,22,53,34,35,36,37,38,39,25,26,27,28,29,30,31,32,33,17,45,47,46,51,24,23,49,48,50,42,43,44,54,2,40,41,13,241,242,224,240,353,342,344,326,236,237,233,238,234,232,231,235]]]],[47,"e",[0,3,684,267,6],["|",[474,537,["*",45]]]],[474,"t",[0,3,687,688,689,690,8,691,692],12],[537,"e",[0,3,474],[",",[]]],[531,"e",[0,3,536,6,1,694],[",",[400,["?",420],["?",396],["?",411],["|",[["*",497],["*",506]]]]]],[400,"e",[0,3],[",",[["?",588],565,["*",["|",[236,237,252,372,272,275,273]]]]]],[420,"e",[0,3,6],[",",[["*",["|",[250,17,238,257,240,234,251,256,232,231,235,244,243,255,241,242,236,237,24,23,233,21,22,252,248,249,246,245,13]]],["*",278],["?",524]]]],[396,"e",[0,3],[",",[["?",195],["*",239],["*",["|",[252,557,372,272,275,276,273,278]]]]]],[497,"e",[0,3,536,6,1],[",",[["|",[400,512]],["?",420],["?",396],["?",411],["|",[["*",497],["*",506]]]]]],[512,"e",[0,3],[",",[["*",459],["?",458],["?",457],["*",["|",[347,287,298,13]]],["?",456],["*",455],["*",288],["*",307],["?",305],["*",281],["*",293],["*",291],["*",395],["?",292],["*",487],["*",297],["?",302],["?",["|",[[",",[["?",[",",[304,["?",290]]]],["?",317]]],301]]],["*",["|",[18,20,19,397,244]]],["?",454],["?",254],["*",399],["*",["|",[21,22]]],["*",263],["*",453],["*",261],["*",452],["*",451],["?",448],["?",415]]]],[506,"e",[0,3,701,6,1],[",",[["|",[400,512]],["?",420],["?",396],["?",411]]]]]]}

  let JATSArchiving11 = deserializeXMLSchema(JATSArchiving11Data,
    '-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD v1.1 20151215//EN',
    'JATS-archivearticle1-mathml3.dtd'
  );

  JATSArchiving11.aliases = [
    '-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD with MathML3 v1.1 20151215//EN'
  ];

  var JATSArchiving12Data = {"literals":["id","xml:lang","xref","xml:base","specific-use","xlink:type","xlink:href","style","xml:space","class","other","href","TEXT","x","content-type","alternatives","definitionURL","encoding","uri","ext-link","email","related-article","related-object","mml:math","tex-math","overline-start","overline-end","roman","sans-serif","sc","strike","underline","underline-start","underline-end","bold","fixed-case","italic","monospace","overline","sup","sub","ruby","inline-graphic","inline-media","named-content","styled-content","private-char","fn","chem-struct","index-term","milestone-start","milestone-end","inline-formula","index-term-range-end","inline-supplementary-material","abbrev","target","hr","mml:ci","mml:semantics","mml:csymbol","mml:grad","mml:divergence","mml:notsubset","mml:cot","mml:sinh","mml:sum","mml:card","mml:notin","mml:in","mml:product","mml:prsubset","mml:subset","mml:cartesianproduct","mml:intersect","mml:setdiff","mml:notprsubset","mml:partialdiff","mml:cosh","mml:tanh","mml:sech","mml:csch","mml:coth","mml:arcsin","mml:arccos","mml:arctan","mml:limit","mml:arccosh","mml:arccot","mml:arccoth","mml:arccsc","mml:arccsch","mml:diff","mml:int","mml:tendsto","mml:factorof","mml:arcsec","mml:cn","mml:arcsech","mml:arcsinh","mml:arctanh","mml:mean","mml:sdev","mml:approx","mml:variance","mml:neq","mml:median","mml:leq","mml:geq","mml:lt","mml:gt","mml:eq","mml:exists","mml:forall","mml:mode","mml:equivalent","mml:implies","mml:not","mml:xor","mml:or","mml:and","mml:lcm","mml:gcd","mml:vector","mml:times","mml:plus","mml:min","mml:max","mml:exp","mml:ceiling","mml:floor","mml:imaginary","mml:real","mml:arg","mml:tan","mml:abs","mml:sec","mml:factorial","mml:root","mml:matrix","mml:rem","mml:power","mml:minus","mml:divide","mml:quotient","mml:compose","mml:lambda","mml:moment","mml:log","mml:ln","mml:image","mml:codomain","mml:domain","mml:ident","mml:inverse","mml:interval","mml:declare","mml:fn","mml:reln","mml:piecewise","mml:apply","mml:sin","mml:csc","mml:union","mml:cos","mml:list","mml:matrixrow","mml:determinant","mml:transpose","mml:selector","mml:vectorproduct","mml:scalarproduct","mml:outerproduct","mml:integers","mml:reals","mml:rationals","mml:naturalnumbers","mml:complexes","mml:cs","mml:cbytes","mml:primes","mml:emptyset","mml:exponentiale","mml:imaginaryi","mml:notanumber","mml:true","mml:false","mml:set","mml:pi","mml:eulergamma","mml:infinity","mml:cerror","mml:share","mml:bind","mml:laplacian","mml:curl","mml:conjugate","mml:malignmark","label","mml:mfrac","mml:msqrt","mml:mroot","mml:maction","mml:mstyle","mml:mi","mml:merror","mml:mpadded","mml:mphantom","mml:mn","mml:mrow","mml:mfenced","mml:menclose","mml:msub","mml:mo","mml:msup","mml:msubsup","mml:munder","mml:mover","mml:maligngroup","mml:munderover","mml:mmultiscripts","mml:mtable","mml:ms","mml:mstack","mml:mtext","mml:mspace","mml:mlongdiv","break","mathcolor","mathbackground","xlink:show","xlink:role","xlink:actuate","xlink:title","graphic","media","p","code","preformat","array","def-list","disp-formula","list","disp-formula-group","title","chem-struct-wrap","rid","supplementary-material","institution","table-wrap","statement","verse-group","address","disp-quote","speech","institution-wrap","permissions","ack","fig","table-wrap-group","boxed-text","fig-group","alt-text","long-desc","position","object-id","abstract","kwd-group","name","year","day","align","vocab","vocab-identifier","attrib","month","season","notes","fn-group","ref-list","country","orientation","glossary","valign","caption","sec","isbn","etal","string-name","string-date","issue","role","volume","toggle","vocab-term-identifier","vocab-term","aff","collab","name-alternatives","conf-date","date","mml:mglyph","count","elocation-id","sec-meta","lpage","fpage","volume-series","volume-id","mml:bvar","mml:domainofapplication","mml:condition","mml:lowlimit","mathvariant","issue-title","mathsize","issue-part","issue-id","dir","supplement","issn-l","issn","width","assigning-authority","conf-name","conf-loc","aff-alternatives","mml:uplimit","publisher-name","fax","pub-id","mml:none","charoff","fontstyle","fontweight","element-citation","mixed-citation","trans-title","nlm-citation","conf-acronym","fontfamily","article-title","city","color","fontsize","phone","background","char","conf-sponsor","page-range","source","collab-alternatives","era","postal-code","iso-8601-date","publisher-loc","state","series","calendar","groupalign","contrib-group","annotation","addr-line","version","comment","trans-source","publication-format","edition","citation-alternatives","person-group","patent","mimetype","part-title","date-in-citation","std","data-title","chapter-title","alt","article-version","mime-subtype","gov","mml:msgroup","bio","mml:msline","mml:msrow","mml:mscarries","type","size","columnalign","rowalign","award-id","funding-source","article-id","superscriptshift","open-access","article-version-alternatives","height","subscriptshift","pub-id-type","rowspan","pub-date","pub-date-not-available","tr","accent","mml:degree","given-names","self-uri","front","cd","crossout","frame","location","notation","shift","funding-group","product","custom-meta-group","suffix","prefix","surname","src","floats-group","journal-id","issue-sponsor","seq","body","degrees","contrib-id","back","lspace","linebreak","depth","on-behalf-of","author-comment","accentunder","rp","rt","mml:mtd","tbody","col","subtitle","page-count","string-conf","conf-theme","conf-num","nested-kwd","compound-kwd","kwd","subj-group","designator","counts","conference","support-group","equalcolumns","trans-abstract","pub-history","history","equalrows","author-notes","title-group","article-categories","textual-form","table","columnspan","def","ext-link-type","framespacing","columnwidth","longdivstyle","glyph-data","publisher-type","publication-type","columnlines","mml:logbase","minlabelspacing","scope","volume-issue-group","close","charspacing","charalign","trans-title-group","open","style-type","style-detail","mml:momentabout","price","rowlines","rowspacing","anonymous","term","selection","form","fence","separator","separators","rspace","stretchy","symmetric","maxsize","minsize","largeop","movablelimits","sub-article","columnspacing","lineleading","linebreakstyle","linebreakmultchar","indentalign","indentshift","indenttarget","indentalignfirst","indentshiftfirst","indentalignlast","indentshiftlast","corresp","response","side","lquote","rquote","stackalign","alignmentscope","resource-name","length","leftoverhang","rightoverhang","mslinethickness","contrib","edge","front-stub","linethickness","numalign","denomalign","bevelled","displaystyle","scriptsizemultiplier","award-group","note","journal-title","journal-subtitle","abbrev-journal-title","glyph-ref","copyright-year","series-title","series-text","list-type","prefix-word","list-content","continued-from","term-head","def-head","def-item","license","subject","compound-subject","list-item","compound-subject-part","time-stamp","access-date","speaker","tfoot","abstract-type","verse-line","kwd-group-type","rationale","ref","ali:free_to_read","date-type","std-organization","colspan","symbol","unstructured-kwd-group","compound-kwd-part","custom-meta","mml:piece","mml:otherwise","meta-name","meta-value","headers","fig-count","table-count","equation-count","ref-count","see","see-also","axis","word-count","thead","name-style","alt-title","colgroup","event-desc","event","article-meta","mml:sep","app-group","app","start_date","article-type","trans-subtitle","license-p","mml:annotation-xml","initials","ali:license_ref","use-type","mml:annotation","mml:mscarry","journal-id-type","pub-type","publisher","abbr","mml:mlabeledtr","mml:mtr","table-wrap-foot","td","arrange","th","rb","span","sig-block","sig","funding-statement","contributed-resource-group","copyright-holder","support-description","resource-group","article","resource-wrap","mml:mprescripts","resource-id","copyright-statement","award-type","support-source","principal-award-recipient","principal-investigator","scriptlevel","source-type","institution-id","scriptminsize","infixlinebreakstyle","decimalpoint","journal-meta","journal-title-group","altimg","index","closure","occurrence","nargs","base","index-type","disp-level","sec-type","object-type","object-id-type","document-type","document-id-type","document-id","source-id-type","source-id","link-type","person-group-type","indent-level","term-type","term-status","ref-type","target-type","fn-type","abbrev-type","support-type","resource-id-type","resource-type","underline-style","veryverythinmathspace","verythinmathspace","thinmathspace","mediummathspace","thickmathspace","verythickmathspace","veryverythickmathspace","voffset","platforms","language-version","language","executable","code-version","code-type","preformat-type","actiontype","display","maxwidth","overflow","order","altimg-width","altimg-height","altimg-valign","alttext","cdgroup","mode","macros","baseline-shift","fig-type","summary","border","rules","cellspacing","cellpadding","event-type","alt-title-type","count-type","product-type","subj-group-type","article-version-type","notes-type","page","vol","related-article-type","currency","authenticated","contrib-id-type","collab-type","description","supplement-type","institution-id-type","fontchar","fontname","format","resolution","x-size","y-size","units","dtd-version","xsi:noNamespaceSchemaLocation","degree-contribution","license-type","deceased","equal-contrib","contrib-type","end_date","response-type"],"schema":[625,[[565,"e",[0,3,14,4,736,596],[",",[]]],[602,"t",[0,3,14,4,596],12],[360,"e",[0,3,14,4],["+",["|",[526,252,296,326,433,382,20,287,19,47,432,291,18,2,13]]]],[526,"e",[0,3,735,514,734,733,246,4,5,6,231,233,230,232],["*",["|",[427,488,297,351,268,298,288,426,252,296,326,433,382,20,287,19,47,432,291,18,2,13]]]],[622,"t",[0,3,14,4,1],["*",["|",[12,248,255,40,39,13]]]],[629,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[541,"t",[0,3,14,4],12],[551,"e",[0,3,732,4,1,5,6,231,233,230,232],["+",["|",[602,599]]]],[599,"t",[0,3,14,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,252,239,260,245,237,258,261,234,235,238,247,249,259,241,243,368,335,336,338,390,391,394,240,242,257,253,254,250,251,485]]]],[256,"e",[0,3],[",",[["*",629],["*",541],["*",622],["*",["|",[565,551]]]]]],[341,"t",[0,3,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[296,"t",[0,3,14,246,4,1],["*",["|",[12,362,342,280,329,248,255,345,353,356,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227,198]]]],[326,"e",[0,3],["+",296]],[299,"t",[0,3,354,358,14,4,1],["*",["|",[12,270,352,275,276,269,34,35,36,37,38,25,26,27,28,29,30,31,32,33,13]]]],[325,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,362,342,280,329,248,255,345,353,356]]]],[324,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[348,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,248,255]]]],[265,"t",[0,3,398,323,14,4],12],[286,"t",[0,3,366,14,4],["*",["|",[12,13]]]],[321,"t",[0,3,607,366,14,4],["*",["|",[12,13]]]],[320,"t",[0,3,4],["*",["|",[12,13]]]],[290,"t",[0,3,14,424,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[317,"t",[0,3,398,323,14,4,1,5,6,231,233,230,232],12],[316,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[423,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[314,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[422,"t",[0,3,606,323,4,1],12],[291,"t",[0,3,14,731,272,273,295,294,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[480,"e",[0,3,14,4,1],[",",[337,["*",598]]]],[598,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[337,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[292,"t",[0,3,424,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[308,"t",[0,3,398,323,14,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[307,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[476,"e",[0,3,14,4,1],[",",[["*",292],["*",308],["?",307],["*",290],["*",317],["*",314],["*",423],["?",316]]]],[488,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[287,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[432,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,248,255]]]],[608,"e",[0,3,14],["+",[",",[328,["?",355]]]]],[328,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,248,255]]]],[355,"t",[0,3,4,1],["*",["|",[12,362,342,280,329,248,255,345,353,356,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[306,"t",[0,3,14,424,4,1],12],[305,"t",[0,3,14,4,1],12],[349,"t",[0,3,14,4,1],12],[387,"t",[0,3,728,4,1],["*",["|",[12,13]]]],[303,"t",[0,3,14,424,4],12],[368,"e",[0,3],["+",["|",[335,336,338]]]],[336,"t",[0,3,471,470,366,603,4,1,5,6,231,233,230,232],["*",["|",[12,54,21,22,57,289,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,198,24,23,55,49,53,51,50,44,45,361,341,376,297,351,364,339,299,325,324,348,375,300,373,270,367,20,303,287,19,306,380,248,255,286,321,320,290,317,316,314,305,275,268,298,265,349,372,370,369,330,355,328,291,276,357,387,350,374,288,319,365,337,18,363,292,308,307,269,47,56,2,40,39,13]]]],[335,"e",[0,3,471,470,366,603,4,1,5,6,231,233,230,232],["+",["|",[54,21,22,57,289,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,198,24,23,55,49,53,51,50,44,45,361,341,376,297,351,364,339,299,325,324,348,375,300,373,270,367,20,303,287,19,306,380,248,255,286,321,320,290,317,316,314,305,275,268,298,265,349,372,370,369,330,355,328,291,276,357,387,350,374,288,319,365,337,18,363,292,308,307,269,47,56,2,40,39,13]]]],[252,"e",[0,3,14,4,1],["*",["|",[362,342,280,329,248,255,345,353,356,20,19,18,198,13]]]],[362,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,342,280,329,248,255,345,353,356]]]],[342,"t",[0,3,14,4,1],12],[280,"t",[0,3,14,280,4,1],["*",["|",[12,13]]]],[20,"t",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[329,"t",[0,3,14,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[248,"t",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[636,"t",[0,3,721,323,272,273,14,4,1],12],[255,"e",[0,3],["*",["|",[248,636]]]],[345,"t",[0,3,14,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[353,"t",[0,3,14,4,1],12],[356,"t",[0,3,14,4,1],12],[18,"t",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[319,"t",[0,3,720,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,360,244]]]],[300,"e",[0,3,566,366,354,358,4],[",",[["?",["|",[[",",[["?",270],["?",275]]],276]]],["?",269],["?",352],["?",289]]]],[270,"t",[0,3,14,4,1],12],[275,"t",[0,3,14,4,1],12],[276,"t",[0,3,14,4,1],12],[352,"t",[0,3,14,4,1],12],[269,"t",[0,3,14,354,358,4,1],12],[289,"t",[0,3,354,358,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,270,352,275,276,269]]]],[351,"e",[0,3],["+",297]],[297,"t",[0,3,718,569,4,1,5,6,231,233,230,232],["*",["|",[12,54,21,22,362,342,280,329,248,255,345,353,356,57,227,360,252,296,326,433,382,20,287,19,47,432,291,18,2,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,40,39,13]]]],[427,"t",[0,3,717,323,716,14,4,1],12],[298,"e",[0,3],["+",["|",[268,288]]]],[268,"e",[0,3,14,587,4,1],[",",[["|",[[",",[419,["?",405]]],405]],["?",418],["?",417]]]],[288,"t",[0,3,14,587,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,426,405,418,419,417]]]],[426,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[405,"t",[0,3,601],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[419,"t",[0,3,601],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[418,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[417,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[19,"t",[0,3,465,323,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[274,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[464,"e",[0,3,246,4,1],["+",236]],[198,"t",[0,3,377,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[485,"t",[0,3,715,14,4,1],["*",["|",[12,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41]]]],[244,"t",[0,3,14,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227,368,335,336,338]]]],[21,"t",[0,3,714,465,713,712,290,303,422,606,4,1,5,6,231,233,230,232],["*",["|",[12,54,21,22,57,227,289,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,422,198,24,23,55,49,53,51,50,44,45,361,341,376,297,351,364,339,299,325,324,348,375,300,373,270,367,20,303,287,19,306,380,248,255,286,321,320,290,317,316,314,305,275,268,298,265,349,372,370,369,330,355,328,291,276,357,387,350,374,288,319,365,337,18,363,292,308,307,269,47,56,2,40,39,13]]]],[618,"t",[0,3,246,14,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227,526,234,235,296,326,287,291,619]]]],[619,"t",[0,3,246,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227,234,235]]]],[257,"e",[0,3,4,14,1],[",",[["?",304],["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]]]]],[382,"e",[0,3,246,14,4,1,5,6,231,233,230,232],[",",[["?",304],["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]]]]],[277,"e",[0,3,711,4,1],[",",[["?",304],["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]]]]],[262,"t",[0,3,14,4,1],12],[263,"t",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[12,13]]]],[416,"e",[0,3],["+",572]],[572,"e",[0,3,4,1,5,6,231,233,230,232],[",",[575,576]]],[575,"t",[0,3],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[576,"t",[0,3],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[15,"e",[0,3],["+",["|",[239,48,237,234,42,43,54,235,238,46,247,462,461,24,23]]]],[461,"t",[0,3,4,1],["*",["|",[12,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,42,43,46,198,24,23,44,45,40,39]]]],[13,"t",[0,3,14,8,4,1],["*",["|",[12,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,42,43,46,44,45,40,39,13]]]],[592,"e",[0,3],[",",[["*",392],["?",["|",[378,395]]],["?",460],["?",459],["*",["|",[360,296,326,13]]],["?",458],["|",[["*",400],["?",401]]],["*",292],["*",308],["?",307],["*",290],["*",317],["*",314],["*",423],["?",316],["*",476],["*",286],["?",319],["?",["|",[[",",[["?",[",",[306,["?",305]]]],["?",349]]],303]]],["*",["|",[20,19,18,415,247]]],["?",456],["?",455],["?",256],["*",406],["*",["|",[21,22]]],["*",266],["*",454],["*",267],["*",414],["*",452],["*",451],["?",450],["?",416]]]],[392,"t",[0,3,398,323,4],12],[378,"t",[0,3,710,449,272,273,295,294,354,14,4,5,6,231,233,230,232],12],[395,"e",[0,3],["+",378]],[460,"e",[0,3],[",",[["*",448],["*",542],["*",543]]]],[448,"e",[0,3,709,272,273,4,1],[",",[["+",["|",[552,553]]],["*",448]]]],[552,"t",[0,3,272,273,295,294,14],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[553,"e",[0,3,272,273,295,294,14],["+",555]],[555,"t",[0,3,14],["*",["|",[12,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,44,45,40,39]]]],[542,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[543,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[458,"e",[0,3,246,4],[",",[["?",198],["?",244],["+",["|",[514,47,236,13]]]]]],[415,"t",[0,3,708,4,1,5,6,231,233,230,232],["*",["|",[12,54,21,22,57,227,289,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,198,24,23,55,49,53,51,50,44,45,485,361,341,376,297,351,364,339,299,325,324,348,375,300,373,270,367,20,303,287,19,306,380,248,255,286,321,320,290,317,316,314,305,275,268,298,265,349,372,370,369,330,355,328,291,276,357,387,350,374,288,319,365,337,18,363,292,308,307,269,47,56,2,40,39,13]]]],[406,"t",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[266,"e",[0,3,560,4,1],[",",[["*",265],["?",304],["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]]]]],[454,"e",[0,3,560,4,1],[",",[["*",265],["?",304],["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]]]]],[267,"e",[0,3,562,272,273,4,1],[",",[["?",198],["?",244],["|",[["+",["|",[447,446,445,13]]],["*",570]]]]]],[447,"t",[0,3,272,273,295,294,14],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[446,"e",[0,3,272,273,295,294,14],["+",571]],[571,"t",[0,3,14],["*",["|",[12,227,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[445,"e",[0,3,272,273,295,294,14],[",",[["+",["|",[447,446]]],["*",445]]]],[570,"t",[0,3,562,272,273,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[514,"t",[0,3,14,4,1],["*",["|",[12,362,342,280,329,248,255,345,353,356,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227,198]]]],[401,"e",[0,3,4],[",",[]]],[400,"e",[0,3,607,366,566,354,358,1],["*",["|",[270,352,275,276,269,289,13]]]],[451,"e",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[299,324,444,325,348,443,339,442,13]]]],[339,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[444,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[443,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[442,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,299,324,444,325,348,443,339,442]]]],[450,"e",[0,3],[",",[["*",302],["?",578],["?",579],["?",580],["?",581],["?",441],["?",585]]]],[302,"e",[0,3,707,302],[",",[]]],[580,"e",[0,3,302],[",",[]]],[578,"e",[0,3,302],[",",[]]],[579,"e",[0,3,302],[",",[]]],[581,"e",[0,3,302],[",",[]]],[441,"e",[0,3,302],[",",[]]],[585,"e",[0,3,302],[",",[]]],[459,"e",[0,3],[",",[341,["*",440],["*",480],["*",588],["?",278]]]],[440,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[588,"t",[0,3,706,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[433,"e",[0,3,14,4,1],[",",[["?",244],["+",236]]]],[456,"t",[0,3],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227,300,289]]]],[590,"t",[0,3,4,1],["*",["|",[12,20,19,18,392,321,320,286,378,395,300,289,400,401]]]],[591,"e",[0,3,705,4,1],[",",[["?",590],["*",392],["?",["|",[378,395]]],["|",[["*",400],["?",401]]],["*",["|",[300,289]]],["*",321],["?",320],["*",286],["?",256],["*",277],["*",406]]]],[455,"e",[0,3],["+",591]],[594,"e",[0,3,14,4,1],[",",[["?",198],["?",244],["*",266],["*",267],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",["|",[595,279]]]]]],[595,"e",[0,3,14,4,1],[",",[["?",304],["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]],["?",256]]]],[278,"e",[0,3,14,4,1],[",",[["?",198],["?",244],["+",["|",[47,13]]]]]],[282,"e",[0,3,14,4,1],[",",[["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",282]]]],[239,"e",[0,3,14,4,281,1],[",",[["?",198],["*",["|",[262,263,20,19,18]]],["|",[["*",["|",[15,234,235]]],438]],["*",["|",[274,256]]]]]],[260,"e",[0,3,264,281,4,1,14],[",",[["*",265],["?",304],["?",198],["?",284],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]],["*",["|",[274,256]]]]]],[245,"e",[0,3,264,281,4,1,14],[",",[["*",265],["?",198],["?",284],["*",266],["*",267],["*",["|",[262,263,20,19,18]]],["+",["|",[15,48,237,234,235,238,461]]],["*",["|",[274,256]]]]]],[48,"t",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[12,262,263,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227,265,198,240,242,239,237,234,235,238]]]],[261,"e",[0,3,264,281,4,1,14],[",",[["?",198],["?",284],["*",266],["*",267],["*",["|",[262,263,20,19,18]]],["*",["|",[258,15,234,235]]]]]],[258,"e",[0,3,264,281,4,1,699],[",",[["*",265],["?",198],["*",284],["*",266],["*",267],["*",["|",[262,263,20,19,18]]],["*",["|",[241,243,245,253,254,250,251,249,236,240,242,15,239,237,234,235,238]]],["*",["|",[274,256]]]]]],[284,"e",[0,3,14,4,1,7],[",",[["?",244],["*",236]]]],[234,"e",[0,3,264,281,4,1,14,379,371,5,6,231,233,230,232],["*",["|",[262,263,266,20,19,18,284,265,267,198,274,256]]]],[43,"t",[0,3,14,371,379,272,273,295,294,4,1,5,6,231,233,230,232],["*",["|",[12,262,263,20,19,18,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,44,45,40,39]]]],[235,"e",[0,3,264,281,4,1,14,371,379,5,6,231,233,230,232],["*",["|",[262,263,266,20,19,18,284,265,267,198,274,256]]]],[42,"e",[0,3,14,4,698,371,379,1,5,6,231,233,230,232],["*",["|",[262,263]]]],[238,"t",[0,3,264,281,4,1,685,8],["*",["|",[12,262,263,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,274,256]]]],[237,"t",[0,3,684,683,682,681,680,679,264,281,4,1,8],["*",["|",[12,20,19,18,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,42,43,46,55,49,53,51,50,44,45,47,56,2,40,39]]]],[247,"e",[0,3,264,281,4,1,14,371,379,5,6,231,233,230,232],[",",[["*",265],["?",198],["*",284],["*",266],["*",267],["*",["|",[262,263,20,19,18]]],["*",["|",[241,243,245,253,254,250,251,249,236,240,242,15,239,237,234,235,238]]],["*",["|",[274,256]]]]]],[259,"e",[0,3,264,281,4,1,14],[",",[["?",198],["?",284],["*",266],["*",267],["*",["|",[262,263,20,19,18]]],["+",249]]]],[249,"e",[0,3,264,281,4,1,14],[",",[["*",265],["?",198],["?",284],["*",266],["*",267],["*",["|",[262,263,20,19,18]]],["*",["|",[253,254,250,251,240,242,15,245,237,241,234,235,238,462]]],["*",["|",[612,274,256]]]]]],[612,"e",[0,3],[",",[["?",198],["?",244],["+",["|",[236,278,47,274,256,13]]]]]],[57,"e",[0,3],[",",[]]],[227,"e",[0,3],[",",[]]],[34,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[35,"t",[0,3,14,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[36,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[37,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[27,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[28,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[29,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[38,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[30,"t",[0,3,293,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[40,"t",[0,3,614,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[39,"t",[0,3,614,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[31,"t",[0,3,293,670,4],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[25,"e",[0,3,4],[",",[]]],[26,"e",[0,3,246,4],[",",[]]],[32,"e",[0,3,4],[",",[]]],[33,"e",[0,3,246,4],[",",[]]],[41,"e",[0,3,14,4],[",",[616,["|",[436,[",",[435,436,435]]]]]]],[616,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[436,"t",[0,3,14,4,1],12],[435,"t",[0,3],12],[414,"e",[0,3,4,1],[",",[["*",535],["*",620],["*",394]]]],[620,"t",[0,3,246,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,390,391,394]]]],[394,"e",[0,3,4,1],["+",236]],[452,"e",[0,3,4,1],[",",[["*",414],["*",621]]]],[621,"e",[0,3,669,4,1],[",",[["*",535],["*",623],["*",624]]]],[623,"e",[0,3,246,4,1],["+",236]],[624,"e",[0,3,4,1],["+",["|",[521,626]]]],[626,"e",[0,3],[",",[521,["*",628]]]],[521,"t",[0,3,14,4,1,5,6,231,233,230,232],["*",["|",[12,34,35,36,37,38,25,26,27,28,29,30,31,32,33,40,39]]]],[628,"t",[0,3,668,272,273,14,4,1],12],[535,"e",[0,3,246,630,4,1,5,6,231,233,230,232],[",",[["|",[["*",391],["*",631]]],["*",390],["*",632],["*",633]]]],[391,"t",[0,3,246,635,280,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,248,255]]]],[390,"t",[0,3,246,630,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[632,"t",[0,3,4,1],["*",["|",[12,427,268,298,248,255,288]]]],[633,"t",[0,3,4,1],["*",["|",[12,427,268,298,288]]]],[631,"t",[0,3,246,667,280,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,248,255]]]],[640,"e",[0,3],[",",[["*",422],["*",641],["*",["|",[360,296,326]]],["*",321],["?",320],["*",286],["?",608],["*",277],["*",406],["?",416]]]],[641,"e",[0,3,14],[",",[["*",537],["*",538],["*",480],["*",539]]]],[537,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[538,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[539,"t",[0,3,666,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[47,"e",[0,3,569,665,4,1],[",",[["?",198],["+",236]]]],[56,"t",[0,3,664,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[2,"t",[0,3,663,377,246,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[54,"t",[0,3,14,371,379,4,1,5,6,231,233,230,232],["*",["|",[12,262,263,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[240,"e",[0,3,544,545,546,547,4,1],[",",[["?",198],["?",244],["?",548],["?",549],["*",["|",[550,13]]],["*",240]]]],[548,"t",[0,3],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[549,"t",[0,3],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[550,"e",[0,3,4,1],[",",[["?",198],["*",489],["*",["|",[464,13]]]]]],[489,"t",[0,3,246,662,661,272,273,295,294,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,241,243,245,239,237,234,235,238]]]],[242,"e",[0,3,544,545,546,547,4,1],[",",[["?",198],["?",244],["+",["|",[554,13]]]]]],[554,"e",[0,3,4,1],[",",[["?",198],["?",244],["+",["|",[236,240,242]]]]]],[52,"t",[0,3,14,4,1],["*",["|",[12,262,263,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[241,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,266,262,263,227,274,256,267,198,239,237,234,235,238]]]],[243,"e",[0,3,14,4,1],[",",[["?",198],["?",284],["*",266],["*",267],["*",["|",[262,263,20,19,18]]],["*",["|",[241,243]]]]]],[24,"t",[0,3,14,4,412,363],12],[338,"e",[0,3,471,470,366,4,1,5,6,231,233,230,232],[",",[["*",["|",[369,297]]],["*",["|",[341,337]]],["?",350],["?",370],["?",365],["?",269],["|",[[",",[["?",275],["?",270],["?",556]]],["?",276]]],["?",557],["?",292],["?",367],["?",324],["?",299],["?",325],["*",["|",[290,319]]],["?",355],["?",328],["*",[",",[["?",306],["?",305]]]],["?",441],["?",357],["*",364],["*",330],["?",361]]]],[236,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,252,239,260,245,237,258,261,234,235,238,247,249,259,241,243,368,335,336,338,390,391,394,240,242,257,253,254,250,251]]]],[253,"e",[0,3,14,4,1],[",",[["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",["|",[274,256]]]]]],[254,"e",[0,3,14,4,1],[",",[558,["+",236]]]],[558,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,426,405,418,419,417]]]],[250,"e",[0,3,14,4,1],[",",[["?",198],["?",244],["*",266],["*",267],["+",["|",[236,250]]],["*",["|",[274,256]]]]]],[251,"e",[0,3,7,483,482,14,4,1],[",",[["?",198],["?",244],["?",440],["+",["|",[561,251]]],["*",["|",[274,256]]]]]],[561,"t",[0,3,660,7,483,482,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[55,"t",[0,3,14,4,377,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,464]]]],[50,"e",[0,3,246,14,563,4,1],[",",[]]],[51,"e",[0,3,246,14,563,4,1],[",",[]]],[44,"t",[0,3,246,377,14,272,273,295,294,4,1,5,6,231,233,230,232],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,252,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,257,253,254,250,251]]]],[45,"t",[0,3,293,7,482,483,377,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,252,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,257,253,254,250,251]]]],[279,"e",[0,3,14,4,1],[",",[["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13,564]]],["*",279]]]],[564,"e",[0,3,14,4,1],[",",[["?",198],["+",["|",[368,335,336,338,536,13]]]]]],[536,"e",[0,3,14,4,1],[",",[["?",198],["+",["|",[236,415]]]]]],[557,"t",[0,3,354,358,14,4],["*",["|",[12,270,352,275,276,269,13]]]],[361,"e",[0,3,14,4,1],["+",236]],[376,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[364,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[375,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,42,43,46,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,44,45,40,39]]]],[373,"t",[0,3,354,358,14,4,1],["*",["|",[12,270,352,275,276,269,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,40,39,13]]]],[367,"t",[0,3,449,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[380,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[372,"t",[0,3,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[370,"t",[0,3,14,280,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[369,"t",[0,3,659,4,1],["*",["|",[12,488,297,351,268,298,288,296,326,287,291,13]]]],[330,"t",[0,3,398,323,4,5,6,231,233,230,232],12],[357,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[374,"t",[0,3,14,4,1],["*",["|",[12,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,42,43,46,44,45,270,275,330,350,567,269,40,39]]]],[567,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,248,255,227]]]],[350,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[556,"t",[0,3,14,4],["*",["|",[12,270,352,275,276,269,13]]]],[365,"t",[0,3,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,227]]]],[363,"t",[0,3,449,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13]]]],[22,"t",[0,3,658,465,657,656,635,655,654,653,265,652,651,14,4,1,5,6,231,233,230,232],["*",["|",[12,54,21,22,57,227,289,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,198,24,23,55,49,53,51,50,44,45,361,341,376,297,351,364,339,299,325,324,348,375,300,373,270,367,20,303,287,19,306,380,248,255,286,321,320,290,317,316,314,305,275,268,298,265,349,372,370,369,330,355,328,291,276,357,387,350,374,288,319,365,337,18,363,292,308,307,269,47,56,2,40,39,13]]]],[421,"e",[0,3],["*",["|",[15,260,245,237,258,261,234,235,238,247,249,259]]]],[285,"e",[0,3,1,650,649,4],[",",[["?",304],["?",198],["?",244],["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["*",["|",[277,278,282,279]]]]]],[304,"e",[0,3],[",",[["*",360],["*",266],["*",267],["?",256]]]],[49,"e",[0,3,648,14,4,1],[",",[489,["|",[49,["*",["|",[582,583]]]]]]]],[53,"e",[0,3,246],[",",[]]],[582,"t",[0,3,246,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,241,243,239,237,234,235,238]]]],[583,"t",[0,3,246,14,4,1],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,241,243,239,237,234,235,238]]]],[97,"t",[6,5,1,8,0,2,9,7,11,10,17,16,386,647],["*",["|",[12,301,593,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[58,"t",[6,5,1,8,0,2,9,7,11,10,17,16,386],["*",["|",[12,301,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[60,"t",[6,5,1,8,0,2,9,7,11,10,17,16,386,408],["*",["|",[12,301,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[160,"e",[6,5,1,8,0,2,9,7,11,10],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]],["*",309],["*",["|",[["*",["|",[310,311,[",",[312,["?",327]]]]]],404,484,473]]],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[193,"e",[6,5,1,8,0,2,9,7,11,10],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]],["*",309],["*",["|",[["*",["|",[310,311,[",",[312,["?",327]]]]]],404,484,473]]],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[192,"e",[6,5,1,8,0,2,9,7,11,10,420],[",",[]]],[191,"e",[6,5,1,8,0,2,9,7,11,10],[",",[60,["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[179,"t",[6,5,1,8,0,2,9,7,11,10,17,16],12],[178,"t",[6,5,1,8,0,2,9,7,11,10,17,16],12],[309,"e",[],["|",[[",",[404,["|",[58,59]]]],[",",[["|",[58,59]],["?",404]]]]]],[593,"e",[],[",",[]]],[310,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[311,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[327,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[312,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[404,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[484,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[473,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[159,"e",[6,5,1,8,0,2,9,7,11,10,17,16],["*",["|",[573,574]]]],[573,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]],[574,"e",[6,5,1,8,0,2,9,7,11,10,17,16],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[158,"e",[],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]],[157,"e",[],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]],[156,"e",[386,475,646,645,17,16],["+",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]],[155,"e",[6,5,1,8,0,2,9,7,11,10,17,16,644],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]],[154,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[153,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[152,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[151,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[150,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[149,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[148,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[147,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[146,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[["*",309],["*",["|",[310,311,[",",[312,["?",327]]]]]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]],[145,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[144,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[143,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[142,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[141,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[140,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[138,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[137,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[135,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[196,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[133,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[132,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[131,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[130,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[129,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[128,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[127,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[126,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[125,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[124,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[122,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[121,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[120,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[119,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[118,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[117,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[116,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[115,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[113,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[112,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[111,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[110,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[109,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[108,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[107,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[105,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[103,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[95,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[94,"e",[6,5,1,8,0,2,9,7,11,10,17,16,386],[",",[]]],[93,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[92,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[77,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[62,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[61,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[195,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[194,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[187,"e",[6,5,1,8,0,2,9,7,11,10,17,16,386],[",",[["*",["*",309]],["*",["*",["|",[310,311,[",",[312,["?",327]]]]]]],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[165,"e",[6,5,1,8,0,2,9,7,11,10,17,16,690],[",",[["*",["*",309]],["*",["*",["|",[310,311,[",",[312,["?",327]]]]]]],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[163,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[74,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[73,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[69,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[68,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[63,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[76,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[75,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[72,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[71,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[67,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[66,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[70,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[86,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[161,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[164,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[134,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[136,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[162,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[64,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[65,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[78,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[79,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[80,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[81,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[82,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[83,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[84,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[85,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[87,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[88,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[89,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[90,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[91,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[96,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[98,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[99,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[100,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[101,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[102,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[104,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[106,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[114,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[123,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[["*",309],["*",["|",[310,311,[",",[312,["?",327]]]]]],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[139,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[["*",309],["*",["|",[310,311,[",",[312,["?",327]]]]]],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[166,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[["*",309],["*",["|",[310,311,[",",[312,["?",327]]]]]],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178]]]]]],[167,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[168,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[169,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[170,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[171,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[172,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[173,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[174,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[175,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[176,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[177,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[180,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[181,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[182,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[183,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[184,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[185,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[186,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[188,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[189,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[190,"e",[6,5,1,8,0,2,9,7,11,10,17,16],[",",[]]],[204,"t",[6,5,1,8,0,2,9,7,11,10,228,229,313,315,318,340,334,333,344,343,346],["*",["|",[12,301,197]]]],[208,"t",[6,5,1,8,0,2,9,7,11,10,228,229,313,315,318,340,334,333,344,343,346],["*",["|",[12,301,197]]]],[213,"t",[6,5,1,8,0,2,9,7,11,10,228,229,313,315,318,340,334,333,344,343,346,491,492,493,429,495,496,497,498,499,500,501,403,430,504,505,506,507,508,509,510,511,512,513],["*",["|",[12,301,197]]]],[224,"t",[6,5,1,8,0,2,9,7,11,10,228,229,313,315,318,340,334,333,344,343,346],["*",["|",[12,301,197]]]],[225,"e",[6,5,1,8,0,2,9,7,11,10,228,229,313,315,318,340,334,333,344,343,346,322,396,431,430],[",",[]]],[222,"t",[6,5,1,8,0,2,9,7,11,10,228,229,313,315,318,340,334,333,344,343,346,517,518],["*",["|",[12,301,197]]]],[301,"e",[6,5,1,8,0,2,9,7,11,10,228,229,420,322,396,283,377,643,313,315,340,334,333,344,343,346],[",",[]]],[383,"e",[6,5,1,8,0,2,9,7,11,10,228,229,264,522,523,524,525],[",",[]]],[331,"e",[6,5,1,8,0,2,9,7,11,10,228,229],[",",[]]],[627,"e",[6,5,1,8,0,2,9,7,11,10,228,229],[",",[]]],[197,"e",[6,5,1,8,0,2,9,7,11,10,228,229,527],[",",[]]],[218,"e",[6,5,1,8,0,2,9,7,11,10,228,229,359],[",",[]]],[209,"e",[6,5,1,8,0,2,9,7,11,10,228,229,318],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[199,"e",[6,5,1,8,0,2,9,7,11,10,228,229,529,530,531,532],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[200,"e",[6,5,1,8,0,2,9,7,11,10,228,229],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[201,"e",[6,5,1,8,0,2,9,7,11,10,228,229],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[203,"e",[6,5,1,8,0,2,9,7,11,10,228,229,634,533,534,637,638,639,403,434,271,520,532,479,478,477,388,472,503,463,467,409,531,431,318,527,453,457,492,491,410,466,359,396,507,510,512,508,511,513,509,500,523,522,430,506,505,504,529,411,468,517,429,315,313,498,474,499,501,525,412,530,481,264,524,389,486,487,399,518,495,490,493,494,413,516,519,496,397,393,497,283,322,340,334,333,344,343,346,671,672,673,674,675,676,677],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[205,"e",[6,5,1,8,0,2,9,7,11,10,228,229],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[206,"e",[6,5,1,8,0,2,9,7,11,10,228,229,396,431,322,429,678],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[207,"e",[6,5,1,8,0,2,9,7,11,10,228,229],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[210,"e",[6,5,1,8,0,2,9,7,11,10,228,229,481,477,494],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[211,"e",[6,5,1,8,0,2,9,7,11,10,228,229,412],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[212,"e",[6,5,1,8,0,2,9,7,11,10,228,229,397],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[214,"e",[6,5,1,8,0,2,9,7,11,10,228,229,393],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[215,"e",[6,5,1,8,0,2,9,7,11,10,228,229,397,393],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[216,"e",[6,5,1,8,0,2,9,7,11,10,228,229,434,271],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[217,"e",[6,5,1,8,0,2,9,7,11,10,228,229,403,271],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[219,"e",[6,5,1,8,0,2,9,7,11,10,228,229,403,434,271],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]]],[220,"e",[6,5,1,8,0,2,9,7,11,10,228,229,397,393],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["*",[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,331]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,331]]]]],["?",[",",[627,["*",[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,331]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,331]]]]]]]]]]],[221,"e",[6,5,1,8,0,2,9,7,11,10,228,229,271,389,388,359,520,467,322,487,503,486,472,410,466,457,453,533,516,474],["*",["|",[611,610]]]],[610,"e",[6,5,1,8,0,2,9,7,11,10,228,229,389,388,359],["+",437]],[611,"e",[6,5,1,8,0,2,9,7,11,10,228,229,389,388,359],["*",437]],[437,"e",[6,5,1,8,0,2,9,7,11,10,228,229,399,463,389,388,359],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[223,"e",[6,5,1,8,0,2,9,7,11,10,228,229,271,519,479,478],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,385,383,384,381]]]],[226,"e",[6,5,1,8,0,2,9,7,11,10,228,229,264,413,468],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,385,383,384,381]],["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,385,383,384,381]],["+",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,385,383,384,381]]]]]],[381,"e",[6,5,1,8,0,2,9,7,11,10,228,229,264,413],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,385,383,384,381]]]],[384,"e",[6,5,1,8,0,2,9,7,11,10,228,229,264],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,331]]]],[385,"e",[6,5,1,8,0,2,9,7,11,10,228,229,264,411,409,534],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,331,605]]]],[605,"e",[6,5,1,8,0,2,9,7,11,10,228,229,411,409],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202,331]]]],[202,"e",[6,5,1,8,0,2,9,7,11,10,228,229,686,490],["+",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[23,"e",[6,5,1,8,0,2,9,7,11,10,687,688,689,642,691,692,693,694,695,696,697,228,229,634,533,534,637,638,639,403,434,271,520,532,479,478,477,388,472,503,463,467,409,531,431,318,527,453,457,492,491,410,466,359,396,507,510,512,508,511,513,509,500,523,522,430,506,505,504,529,411,468,517,429,315,313,498,474,499,501,525,412,530,481,264,524,389,486,487,399,518,495,490,493,494,413,516,519,496,397,393,497,283,322],["*",["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]]]],[604,"t",[6,5,1,8,0,2,9,7,11,10,408,268,17,16,420],12],[600,"e",[6,5,1,8,0,2,9,7,11,10,408,268,17,16,420],["+",236]],[59,"e",[6,5,1,8,0,2,9,7,11,10,17,16,408,268],[",",[["|",[159,158,157,156,155,154,153,152,151,150,149,148,147,146,145,144,143,142,141,140,138,137,135,196,133,132,131,130,129,128,127,126,125,124,122,121,120,119,118,117,116,115,113,112,111,110,109,108,107,105,103,95,94,93,92,77,62,61,195,194,187,165,163,74,73,69,68,63,76,75,72,71,67,66,70,86,161,164,134,136,162,64,65,78,79,80,81,82,83,84,85,87,88,89,90,91,96,98,99,100,101,102,104,106,114,123,139,166,167,168,169,170,171,172,173,174,175,176,177,180,181,182,183,184,185,186,188,189,190,59,97,58,60,160,193,192,191,179,178,204,208,213,224,225,222,218,197,209,199,200,201,203,205,206,207,210,211,212,214,215,216,217,219,220,221,223,226,202]],["*",["|",[604,600]]]]]],[462,"e",[0,3,14,7,700,322,701,410,702,703,704,4],[",",[["|",[["*",439],["*",589]]],["|",[[",",[["?",586],["?",559],["+",438]]],["+",402]]]]]],[586,"e",[0,3,14,7,271,347,332,283],["+",402]],[559,"e",[0,3,14,7,271,347,332,283],["+",402]],[438,"e",[0,3,14,7,271,347,332,283],["+",402]],[589,"e",[0,3,14,7,617,322,271,347,332,283],["*",439]],[439,"e",[0,3,14,7,617,322,271,347,332,283],[",",[]]],[402,"e",[0,3,14,7,271,347,332,283],["+",["|",[615,613]]]],[615,"t",[0,3,14,7,609,584,577,475,399,568,271,347,332,283],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,241,243,227,245,368,335,336,338,240,242,236,239,237,234,235,238]]]],[613,"t",[0,3,14,7,609,584,577,475,399,568,271,347,332,283],["*",["|",[12,20,19,18,54,21,22,57,34,35,36,37,38,25,26,27,28,29,30,31,32,33,41,15,42,43,46,48,52,24,23,55,49,53,51,50,44,45,47,56,2,40,39,13,241,243,227,245,368,335,336,338,240,242,236,239,237,234,235,238]]]],[46,"e",[0,3,719,268,4],["|",[469,540,["*",42]]]],[469,"t",[0,3,722,723,724,725,8,726,727],12],[540,"e",[0,3,469],[",",[]]],[625,"e",[0,3,597,4,1,729,730],[",",[407,["?",425],["?",428],["?",421],["|",[["*",502],["*",515]]]]]],[407,"e",[0,3],[",",[["?",640],592,["*",["|",[240,242,257,382,278,282,277]]]]]],[425,"e",[0,3,4],[",",[["*",["|",[252,15,239,260,245,237,258,261,234,235,238,247,249,259,241,243,240,242,24,23,236,21,22,257,253,254,250,251,13]]],["*",285],["?",618]]]],[428,"e",[0,3],[",",[["?",198],["*",244],["*",["|",[257,594,382,278,282,279,277,285]]]]]],[502,"e",[0,3,597,4,1],[",",[["|",[407,528]],["?",425],["?",428],["?",421],["|",[["*",502],["*",515]]]]]],[528,"e",[0,3],[",",[["*",392],["?",["|",[378,395]]],["?",460],["?",459],["*",["|",[360,296,326,13]]],["?",458],["|",[["*",400],["?",401]]],["*",292],["*",308],["?",307],["*",290],["*",317],["*",314],["*",423],["?",316],["*",476],["*",286],["?",319],["?",["|",[[",",[["?",[",",[306,["?",305]]]],["?",349]]],303]]],["*",["|",[20,19,18,415,247]]],["?",456],["?",455],["?",256],["*",406],["*",["|",[21,22]]],["*",266],["*",454],["*",267],["*",414],["*",452],["*",451],["?",450],["?",416]]]],[515,"e",[0,3,737,4,1],[",",[["|",[407,528]],["?",425],["?",428],["?",421]]]]]]}

  let JATSArchiving12 = deserializeXMLSchema(JATSArchiving12Data,
    '-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD v1.2 20190208//EN',
    'JATS-archivearticle1-mathml3.dtd'
  );

  JATSArchiving12.aliases = [
    '-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD with MathML3 v1.2 20190208//EN'
  ];

  // EXPERIMENTAL: at the moment, we do not necessarily need this
  // as a Texture plugin. We only use it for development, to be able to
  // validate output generated by Texture against the particular JATS DTDs.
  // When this plugin is loaded, any JATS file will be validated during loading.
  texture.Texture.registerPlugin({
    name: 'jats-plugin',
    description: 'Provides JATS validators',
    configure (configurator) {
      let articleConfig = configurator.getConfiguration('article');
      // ATTENTION: Texture does not support JATS 1.0, thus I have removed the validator here
      _registerValidator(articleConfig, JATSArchiving11);
      _registerValidator(articleConfig, JATSArchiving12);
    }
  });

  function _registerValidator (articleConfig, schema) {
    articleConfig.registerSchemaId(schema.publicId);
    let validator = {
      schemaId: schema.publicId,
      validate (xmlDom) {
        return validateXML(schema, xmlDom)
      }
    };
    articleConfig.addValidator(schema.publicId, validator);
    if (schema.aliases) {
      schema.aliases.forEach(alias => {
        articleConfig.addValidator(alias, validator);
      });
    }
  }

})));

//# sourceMappingURL=./texture-plugin-jats.js.map